from __future__ import division

import functools
from Ganga.Utility.logging import getLogger

from Ganga.Core import GangaException
from Ganga.Core.GangaRepository import InaccessibleObjectError, RepositoryError

import time
import threading

from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
from Ganga.GPIDev.Base.Objects import GangaObject
from Ganga.GPIDev.Schema import Schema, Version
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
from Ganga.Utility.Config import getConfig

logger = getLogger()

_reg_id_str = '_registry_id'
_id_str = 'id'

class RegistryError(GangaException):

    def __init__(self, what=''):
        super(RegistryError, self).__init__(self, what)
        self.what = what

    def __str__(self):
        return "RegistryError: %s" % self.what


class RegistryAccessError(RegistryError):

    """ This error is raised if the request is valid in principle, 
        but the Registry cannot be accessed at the moment."""

    def __init__(self, what=''):
        super(RegistryAccessError, self).__init__(what)

    def __str__(self):
        return "RegistryAccessError: %s" % self.what


class RegistryLockError(RegistryError):

    """ This error is raised if the request is valid in principle,
        but the object is locked by another Ganga session"""

    def __init__(self, what=''):
        super(RegistryLockError, self).__init__(what)

    def __str__(self):
        return "RegistryLockError: %s" % self.what


class ObjectNotInRegistryError(RegistryError):

    """ This error is raised if an object has been associated to this registry,
        but is not actually in the registry. This most probably indicates an internal Ganga error."""

    def __init__(self, what=''):
        super(ObjectNotInRegistryError, self).__init__(what)

    def __str__(self):
        return "ObjectNotInRegistryError: %s" % self.what


class RegistryKeyError(RegistryError, KeyError):

    """ This error is raised if the given id is not found in the registry """

    def __init__(self, what=''):
        super(RegistryKeyError, self).__init__(what)

    def __str__(self):
        return "RegistryKeyError: %s" % self.what


class RegistryIndexError(RegistryError, IndexError):

    """ This error is raised if the given id is not found in the registry """

    def __init__(self, what=''):
        super(RegistryIndexError, self).__init__(what)

    def __str__(self):
        return "RegistryIndexError: %s" % self.what


def makeRepository(registry):
    """Factory that selects, imports and instantiates the correct GangaRepository"""
    if registry.type in ["LocalXML", "LocalPickle"]:
        from Ganga.Core.GangaRepository.GangaRepositoryXML import GangaRepositoryLocal
        return GangaRepositoryLocal(registry)
    elif registry.type in ["SQLite"]:
        from Ganga.Core.GangaRepository.GangaRepositorySQLite import GangaRepositorySQLite
        return GangaRepositorySQLite(registry)
    elif registry.type in ["Transient"]:
        from Ganga.Core.GangaRepository.GangaRepository import GangaRepository
        return GangaRepository(registry)
    elif registry.type in ["ImmutableTransient"]:
        from Ganga.Core.GangaRepository.GangaRepositoryImmutableTransient import GangaRepositoryImmutableTransient
        return GangaRepositoryImmutableTransient(registry, registry.location, registry.file_ext, registry.pickle_files)
    else:
        raise RegistryError("Repository %s: Unknown repository type %s" % (registry.name, registry.type))


class IncompleteObject(GangaObject):

    """ This class represents an object that could not be loaded on startup"""

    _schema = Schema(Version(0, 0), {})
    _name = "IncompleteObject"
    _category = "internal"
    _hidden = 1

    _exportmethods = ['reload', 'remove', '__repr__']

    def __init__(self, registry, this_id):
        super(IncompleteObject, self).__init__()
        self.registry = registry
        self.id = this_id

    def reload(self):
        self.registry._lock.acquire()
        try:

            if self.registry.checkShouldFlush():
                self.registry.repository.flush([self.registry._objects[self.id]])
                self.registry._load([self.id])
            if self.id not in self.registry_loaded_ids:
                self.registry._load([self.id])
                self.registry._loaded_ids.append(self.id)
            logger.debug("Successfully reloaded '%s' object #%i!" % (self.registry.name, self.id))
            for d in self.registry.changed_ids.itervalues():
                d.add(self.id)
        finally:
            self.registry._lock.release()

    def remove(self):
        self.registry._lock.acquire()
        try:
            if len(self.registry.repository.lock([self.id])) == 0:
                errstr = "Could not lock '%s' object #%i!" % (self.registry.name, self.id)
                try:
                    errstr += " Object is locked by session '%s' " % self.registry.repository.get_lock_session(self.id)
                except Exception as err:
                    logger.debug("Remove Lock error: %s" % str(err))
                raise RegistryLockError(errstr)
            self.registry.repository.delete([self.id])
            for d in self.registry.changed_ids.itervalues():
                d.add(self.id)
        finally:
            self.registry._lock.release()

    def __repr__(self):
        return "Incomplete object in '%s', ID %i. Try reload() or remove()." % (self.registry.name, self.id)


def synchronised(f):
    """
    This decorator must be attached to a method on a ``Registry``
    It uses the object's lock to make sure that the object is held for the duration of the decorated function
    """
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._lock:
            return f(self, *args, **kwargs)
    return decorated


class RegistryFlusher(threading.Thread):
    """
    This class is intended to be used by the registry to perfom
    automatic flushes on a fixed schedule so that information is not
    lost if Ganga is shut down abruptly.
    """
    def __init__(self, registry, *args, **kwargs):
        super(RegistryFlusher, self).__init__(*args, **kwargs)
        self.registry = registry
        self._stop = threading.Event()

    def stop(self):
        """
        Ask the thread to stop what it is doing and it will finish
        the next chance it gets.
        """
        self._stop.set()

    @property
    def stopped(self):
        return self._stop.isSet()

    def join(self, *args, **kwargs):
        self.stop()
        super(RegistryFlusher, self).join(*args, **kwargs)

    def run(self):
        """
        This will run an indefinite loop which periodically checks
        whether it should stop. In between calls to ``flush_all`` it
        will wait for a fixed period of time.
        """
        sleep_period = getConfig('Registry')['AutoFlusherWaitTime']
        sleeps_per_second = 10  # This changes the granularity of the sleep.
        while not self.stopped:
            for i in range(sleep_period*sleeps_per_second):
                time.sleep(1/sleeps_per_second)
                if self.stopped:
                    return
            # We want this to be a non-blocking lock to avoid this
            # interfering with interactive work or monitoring. It
            # will try again in a while anyway.
            if self.registry._lock.acquire(blocking=False):
                try:
                    logger.debug('Auto-flushing', self.registry.name)
                    self.registry.flush_all()
                finally:
                    self.registry._lock.release()


class Registry(object):

    """Ganga Registry
    Base class providing a dict-like locked and lazy-loading interface to a Ganga repository
    """

    def __init__(self, name, doc, dirty_flush_counter=10, update_index_time=30, dirty_max_timeout=60, dirty_min_timeout=30):
        """Registry constructor, giving public name and documentation"""
        self.name = name
        self.doc = doc
        self._hasStarted = False
        self.dirty_flush_counter = dirty_flush_counter
        self.dirty_hits = 0
        self.update_index_time = update_index_time
        self._update_index_timer = 0
        self._needs_metadata = False
        self.metadata = None
        self._lock = threading.RLock()
        self.hard_lock = {}
        self.changed_ids = {}
        self._autoFlush = True

        self._loaded_ids = []

        self._parent = None

        self.repository = None
        self._objects = None
        self._incomplete_objects = None

        ## Record the last dirty and flush times to determine whether an idividual flush command should flush
        ## Logc to use these is implemented in checkShouldFlush()
        self._dirtyModTime = None
        self._flushLastTime = None
        self._dirty_max_timeout = dirty_max_timeout
        self._dirty_min_timeout = dirty_min_timeout


        self._inprogressDict = {}


        self.shouldReleaseRun = True

        self.flush_thread = None

#        self.releaseThread = threading.Thread(target=self.trackandRelease, args=())
#        self.releaseThread.daemon = True
#        self.releaseThread.start()

    def hasStarted(self):
        return self._hasStarted

    def lock_transaction(self, this_id, action):
        while this_id in self._inprogressDict.keys():
            logger.debug("Getting item being operated on: %s" % this_id)
            logger.debug("Currently in state: %s" % self._inprogressDict[this_id])
            #import traceback
            #traceback.print_stack()
            #import sys
            #sys.exit(-1)
            #time.sleep(0.05)
        self._inprogressDict[this_id] = action
        if this_id not in self.hard_lock.keys():
            self.hard_lock[this_id] = threading.Lock()
        self.hard_lock[this_id].acquire()

    def unlock_transaction(self, this_id):
        self._inprogressDict[this_id] = False
        del self._inprogressDict[this_id]
        self.hard_lock[this_id].release()

    # Methods intended to be called from ''outside code''
    def __getitem__(self, this_id):
        """ Returns the Ganga Object with the given id.
            Raise RegistryKeyError"""
        logger.debug("__getitem__")
        try:
            return self._objects[this_id]
        except KeyError as err:
            logger.debug("Repo KeyError: %s" % err)
            logger.debug("Keys: %s id: %s" % (self._objects.keys(), this_id))
            if this_id in self._incomplete_objects:
                return IncompleteObject(self, this_id)
            raise RegistryKeyError("Could not find object #%s" % this_id)

    @synchronised
    def __len__(self):
        """ Returns the current number of root objects """
        logger.debug("__len__")
        return len(self._objects)

    @synchronised
    def __contains__(self, this_id):
        """ Returns True if the given ID is in the registry """
        logger.debug("__contains__")
        return this_id in self._objects

    def updateLocksNow(self):
        logger.debug("updateLocksNow")
        self.repository.updateLocksNow()

    def trackandRelease(self):

        while self.shouldReleaseRun is True:

            ## Needed import for shutdown
            import time
            timeNow = time.time()

            modTime = self._dirtyModTime
            if modTime is None:
                modTime = timeNow
            dirtyTime = self._flushLastTime
            if dirtyTime is None:
                dirtyTime = timeNow

            delta_1 = abs(timeNow - modTime)
            delta_2 = abs(timeNow - dirtyTime)

            if delta_1 > self._dirty_max_timeout and delta_2 > self._dirty_max_timeout:

                 flush_thread = threading.Thread(target=self._flush, args=())
                 flush_thread.run()
    
            time.sleep(0.5)

    def turnOffAutoFlushing(self):
        self._autoFlush = False

    def turnOnAutoFlushing(self):
        self._autoFlush = True

    def isAutoFlushEnabled(self):
        return self._autoFlush

    def checkDirtyFlushtimes(self, timeNow):
        self._dirtyModTime = timeNow
        if self._flushLastTime is None:
            self._flushLastTime = timeNow

    @synchronised
    def checkShouldFlush(self):
        logger.debug("checkShouldFlush")

        timeNow = time.time()

        self.checkDirtyFlushtimes(timeNow)

        timeDiff = (self._dirtyModTime - self._flushLastTime)

        if timeDiff > self._dirty_min_timeout:
            hasMinTimedOut = True
        else:
            hasMinTimedOut = False

        if timeDiff > self._dirty_max_timeout:
            hasMaxTimedOut = True
        else:
            hasMaxTimedOut = False

        if self.dirty_hits > self.dirty_flush_counter:
            countLimitReached = True
        else:
            countLimitReached = False

        # THIS IS THE MAIN DECISION ABOUT WHETHER TO FLUSH THE OBJECT TO DISK!!!
        # if the minimum amount of time has passed __AND__ we meet a sensible condition for wanting to flush to disk
        decision = hasMinTimedOut and (hasMaxTimedOut or countLimitReached)
       
        ## This gives us the ability to automatically turn off the automatic flushing externally if required
        decision = decision and self._autoFlush

        ## Can't autosave if a flush is in progress. Wait until next time.
        if len(self._inprogressDict.keys()) != 0:
            decision = False

        if decision is True:
            self._flushLastTime = timeNow
            self.dirty_hits = 0

        return decision

    def _getObjects(self):
        logger.debug("_getObjects")
        returnable = self._objects
        return returnable

    @synchronised
    def ids(self):
        """ Returns the list of ids of this registry """
        logger.debug("ids")
        if self.hasStarted() is True and\
                (time.time() > self._update_index_timer + self.update_index_time):
            try:
                changed_ids = self.repository.update_index()
                for this_d in self.changed_ids.itervalues():
                    this_d.update(changed_ids)
            except Exception as err:
                pass
            self._update_index_timer = time.time()

        return sorted(self._objects.keys())

    @synchronised
    def items(self):
        """ Return the items (ID,obj) in this registry. 
        Recommended access for iteration, since accessing by ID can fail if the ID iterator is old"""
        logger.debug("items")
        if self.hasStarted() is True and\
                (time.time() > self._update_index_timer + self.update_index_time):
            try:
                changed_ids = self.repository.update_index()
                for this_d in self.changed_ids.itervalues():
                    this_d.update(changed_ids)
            except Exception as err:
                pass

            self._update_index_timer = time.time()

        return sorted(self._objects.items())

    def iteritems(self):
        """ Return the items (ID,obj) in this registry."""
        logger.debug("iteritems")
        returnable = self.items()
        return returnable

    def _checkObjects(self):
        for key, _obj in self._objects.iteritems():
            summary = "found: "
            try:
                if hasattr(_obj, _id_str):
                    summary = summary + "%s = '%s'" % (_id_str, getattr(_obj, _id_str))
                    assert(getattr(_obj, _id_str) == key)
                if hasattr(_obj, _reg_id_str):
                    summary = summary + " %s = '%s'" % (_reg_id_str, getattr(_obj, _reg_id_str))
                    assert(getattr(_obj, _reg_id_str) == key)
            except:
                logger.error(summary)
                raise
        return

    @synchronised
    def keys(self):
        """ Returns the list of ids of this registry """
        logger.debug("keys")
        returnable = self.ids()
        return returnable

    @synchronised
    def values(self):
        """ Return the objects in this registry, in order of ID.
        Besides items() this is also recommended for iteration."""
        logger.debug("values")
        returnable = [it[1] for it in self.items()]
        return returnable

    def __iter__(self):
        logger.debug("__iter__")
        returnable = iter(self.values())
        return returnable

    def find(self, _obj):
        """Returns the id of the given object in this registry, or 
        Raise ObjectNotInRegistryError if the Object is not found"""

        obj = stripProxy(_obj)
        try:
            if hasattr(obj, _reg_id_str):
                obj_reg_id = getattr(obj, _reg_id_str)
                objects_obj = self._objects[obj_reg_id]
                assert obj == objects_obj
                if hasattr(obj, _id_str):
                    if hasattr(objects_obj, _id_str):
                        assert getattr(obj, _id_str) == getattr(objects_obj, _id_str)
                assert obj_reg_id == getattr(objects_obj, _reg_id_str)
                return obj_reg_id
            elif hasattr(obj, _id_str):
                obj_id = getattr(obj, _id_str)
                objects_obj = self._objects[obj_id]
                if hasattr(objects_obj, _reg_id_str):
                    assert obj_id == getattr(objects_obj, _reg_id_str)
                assert obj_id == getattr(objects_obj, _id_str)
                return obj_id
            else:
                raise ObjectNotInRegistryError("Repo find: Object '%s' does not seem to be in this registry: %s !" % (getName(obj), self.name))
        except AttributeError as err:
            logger.debug("%s" % str(err))
            raise ObjectNotInRegistryError("Object %s does not seem to be in any registry!" % getName(obj))
        except AssertionError as err:
            logger.warning("%s" % str(err))
            raise ObjectNotInRegistryError("Object '%s' is a duplicated version of the one in this registry: %s !" % (getName(obj), self.name))
        except KeyError as err:
            logger.debug("%s", str(err))
            raise ObjectNotInRegistryError("Object '%s' does not seem to be in this registry: %s !" % (getName(obj), self.name))

    @synchronised
    def clean(self, force=False):
        """Deletes all elements of the registry, if no other sessions are present.
        if force == True it removes them regardless of other sessions.
        Returns True on success, False on failure."""
        logger.debug("clean")
        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot clean a disconnected repository!")
        try:
            if not force:
                other_sessions = self.repository.get_other_sessions()
                if len(other_sessions) > 0:
                    logger.error("The following other sessions are active and have blocked the clearing of the repository: \n * %s" % ("\n * ".join(other_sessions)))
                    return False
            self.repository.reap_locks()
            self.repository.delete(self._objects.keys())
            self.dirty_hits = 0
            self.changed_ids = {}
            self.repository.clean()
        except (RepositoryError, RegistryAccessError, RegistryLockError, ObjectNotInRegistryError) as err:
            raise err
        except Exception as err:
            logger.debug("Clean Unknown Err: %s" % str(err))
            raise err

    @synchronised
    def __safe_add(self, obj, force_index=None):
        logger.debug("__safe_add")
        if force_index is None:
            ids = self.repository.add([obj])
        else:
            if len(self.repository.lock([force_index])) == 0:
                raise RegistryLockError("Could not lock '%s' id #%i for a new object!" % (self.name, force_index))
            ids = self.repository.add([obj], [force_index])

        obj._setRegistry(self)
        obj._registry_locked = True

        this_id = self.find(obj)
        try:
            self.lock_transaction(this_id, "_add")

            self.repository.flush(ids)
            for this_v in self.changed_ids.itervalues():
                this_v.update(ids)

            for _id in ids:
                if hasattr(self._objects[_id], _reg_id_str):
                    assert(getattr(self._objects[_id], _reg_id_str) == _id)
                if hasattr(self._objects[_id], _id_str):
                    assert(getattr(self._objects[_id], _id_str) == _id)

            logger.debug("_add-ed as: %s" % str(ids))
        finally:
            self.unlock_transaction(this_id)
        return ids[0]

    # Methods that can be called by derived classes or Ganga-internal classes like Job
    # if the dirty objects list is modified, the methods must be locked by self._lock
    # all accesses to the repository must also be locked!

    @synchronised
    def _add(self, _obj, force_index=None):
        """ Add an object to the registry and assigns an ID to it. 
        use force_index to set the index (for example for metadata). This overwrites existing objects!
        Raises RepositoryError"""
        logger.debug("_add")
        obj = stripProxy(_obj)

        if self.hasStarted() is not True:
            raise RepositoryError("Cannot add objects to a disconnected repository!")

        this_id = None
        returnable_id = None

        try:
            returnable_id = self.__safe_add(obj, force_index)
            self._loaded_ids.append(returnable_id)
        except (RepositoryError) as err:
            raise err
        except Exception as err:
            logger.debug("Unknown Add Error: %s" % str(err))
            raise err

        self._updateIndexCache(obj)

        return returnable_id

    @synchronised
    def _remove(self, _obj, auto_removed=0):
        """ Private method removing the obj from the registry. This method always called.
        This method may be overriden in the subclass to trigger additional actions on the removal.
        'auto_removed' is set to true if this method is called in the context of obj.remove() method to avoid recursion.
        Only then the removal takes place. In the opposite case the obj.remove() is called first which eventually calls
        this method again with "auto_removed" set to true. This is done so that obj.remove() is ALWAYS called once independent
        on the removing context.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise RegistryLockError
        Raise ObjectNotInRegistryError"""
        logger.debug("_remove")
        obj = stripProxy(_obj)
        try:
            self.__reg_remove(obj, auto_removed)
        except ObjectNotInRegistryError as err:
            try:
                ## Actually  make sure we've removed the object from the repo 
                if hasattr(obj, _reg_id_str):
                    del self._objects[getattr(obj, _reg_id_str)]
            except Exception as err:
                pass
            pass
        except Exception as err:
            raise err

    def __reg_remove(self, obj, auto_removed=0):

        logger.debug("_reg_remove")
        u_id = self.find(obj)

        obj_id = id(obj)

        try:
            self.lock_transaction(obj_id, "_remove")


            if self.hasStarted() is not True:
                raise RegistryAccessError("Cannot remove objects from a disconnected repository!")
            if not auto_removed and hasattr(obj, "remove"):
                obj.remove()
            else:
                this_id = self.find(obj)
                try:
                    self._write_access(obj)
                except RegistryKeyError as err:
                    logger.debug("Registry KeyError: %s" % str(err))
                    logger.warning("double delete: Object #%i is not present in registry '%s'!" % (this_id, self.name))
                    return
                logger.debug('deleting the object %d from the registry %s', this_id, self.name)
                try:
                    self.repository.delete([this_id])
                    del obj
                    for this_v in self.changed_ids.itervalues():
                        this_v.add(this_id)
                except (RepositoryError, RegistryAccessError, RegistryLockError) as err:
                    raise err
                except Exception as err:
                    logger.debug("unknown Remove Error: %s" % str(err))
                    raise err
        finally:

            self.unlock_transaction(obj_id)

    @synchronised
    def _flush(self, objs):
        """
        Flush a set of objects to the persistency layer immediately

        Only those objects passed in will be flushed and only if they are dirty.

        Args:
            objs: a list of objects to flush
        """
        logger.debug("_flush")

        if isType(objs, (list, tuple, GangaList)):
            objs = [stripProxy(_obj) for _obj in objs]
        else:
            objs = [stripProxy(objs)]

        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot flush to a disconnected repository!")

        for obj in objs:
            # check if the object is dirty, if not do nothing
            if not obj._dirty:
                continue

            with obj.const_lock:
                # flush the object. Need to call _getWriteAccess for consistency reasons
                # TODO: getWriteAccess should only 'get write access', as that's not needed should it be called here?
                obj._getWriteAccess()
                obj_id = getattr(obj, _reg_id_str)
                self.repository.flush([obj_id])
                obj._setFlushed()

    @synchronised
    def flush_all(self):
        """
        This will attempt to flush all the jobs in the registry.
        It does this via ``_flush`` so the same conditions apply.
        """
        if self.hasStarted():
            self._flush(self.values())

        if self.metadata and self.metadata.hasStarted():
            self.metadata.flush_all()

    def _read_access(self, _obj, sub_obj=None):
        """Obtain read access on a given object.
        sub-obj is the object the read access is actually desired (ignored at the moment)
        Raise RegistryAccessError
        Raise RegistryKeyError"""
        logger.debug("_read_access")
        obj_id = id(stripProxy(_obj))
        if obj_id in self._inprogressDict.keys():
            return

        with _obj.const_lock:
            self.__safe_read_access(_obj, sub_obj)

    @synchronised
    def _updateIndexCache(self, _obj):
        logger.debug("_updateIndexCache")
        obj = stripProxy(_obj)
        if self.find(obj) in self._inprogressDict.keys():
            return

        self.repository.updateIndexCache(obj)

    @synchronised
    def _load(self, obj_ids):
        logger.debug("_load")
        these_ids = []
        for obj_id in obj_ids:
            this_id = id(self[obj_id])
            these_ids.append(this_id)
            self.lock_transaction(this_id, "_load")

        try:
            for obj_id in obj_ids:
                self.repository.load([obj_id])
        except Exception as err:
            logger.error("Error Loading Jobs!")
            raise err
        finally:
            for obj_id in these_ids:
                self.unlock_transaction(obj_id)

    def __safe_read_access(self,  _obj, sub_obj):
        logger.debug("_safe_read_access")
        obj = stripProxy(_obj)
        if self.find(obj) in self._inprogressDict.keys():
            return

        if self.hasStarted() is not True:
            raise RegistryAccessError("The object #%i in registry '%s' is not fully loaded and the registry is disconnected! Type 'reactivate()' if you want to reconnect." % (self.find(obj), self.name))

        if hasattr(obj, "_registry_refresh"):
            delattr(obj, "_registry_refresh")
        assert not hasattr(obj, "_registry_refresh")

        try:
            this_id = self.find(obj)
            try:
                if this_id not in self._loaded_ids:
                    self._load([this_id])
                    self._loaded_ids.append(this_id)
            except KeyError as err:
                logger.error("_read_access KeyError %s" % str(err))
                raise RegistryKeyError("Read: The object #%i in registry '%s' was deleted!" % (this_id, self.name))
            except InaccessibleObjectError as err:
                raise RegistryKeyError("Read: The object #%i in registry '%s' could not be accessed - %s!" % (this_id, self.name, str(err)))
            #finally:
            #    pass
            for this_d in self.changed_ids.itervalues():
                this_d.add(this_id)
        except (RepositoryError, RegistryAccessError, RegistryLockError, ObjectNotInRegistryError) as err:
            raise err
        except Exception as err:
            logger.debug("Unknown read access Error: %s" % str(err))
            raise err
        #finally:
        #    pass

    def _write_access(self, _obj):
        """Obtain write access on a given object.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise RegistryLockError
        Raise ObjectNotInRegistryError (via self.find())"""
        logger.debug("_write_access")
        obj = stripProxy(_obj)
        obj_id = id(_obj)
        if obj_id in self._inprogressDict.keys():
            return

        with obj.const_lock:
            self.__write_access(obj)

    def __write_access(self, _obj):
        logger.debug("__write_acess")
        obj = stripProxy(_obj)
        this_id = self.find(obj)
        if this_id in self._inprogressDict.keys():
            for this_d in self.changed_ids.itervalues():
                this_d.add(this_id)
            return

        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot get write access to a disconnected repository!")
        if not hasattr(obj, '_registry_locked') or not obj._registry_locked:
            try:
                this_id = self.find(obj)
                try:
                    if len(self.repository.lock([this_id])) == 0:
                        errstr = "Could not lock '%s' object #%i!" % (self.name, this_id)
                        try:
                            errstr += " Object is locked by session '%s' " % self.repository.get_lock_session(this_id)
                        except RegistryLockError as err:
                            raise err
                        except Exception as err:
                            logger.debug( "Unknown Locking Exception: %s" % str(err) )
                            raise err
                        raise RegistryLockError(errstr)
                except (RepositoryError, RegistryAccessError, RegistryLockError, ObjectNotInRegistryError) as err:
                    raise err
                except Exception as err:
                    logger.debug("Unknown write access Error: %s" % str(err))
                    raise err
                finally:  # try to load even if lock fails
                    try:
                        if this_id not in self._loaded_ids:
                            self._load([this_id])
                            self._loaded_ids.append(this_id)
                            if hasattr(obj, "_registry_refresh"):
                                delattr(obj, "_registry_refresh")
                    except KeyError, err:
                        logger.debug("_write_access KeyError %s" % str(err))
                        raise RegistryKeyError("Write: The object #%i in registry '%s' was deleted!" % (this_id, self.name))
                    except InaccessibleObjectError as err:
                        raise RegistryKeyError("Write: The object #%i in registry '%s' could not be accessed - %s!" % (this_id, self.name, str(err)))
                    #finally:
                    #    pass
                    for this_d in self.changed_ids.itervalues():
                        this_d.add(this_id)
                obj._registry_locked = True
            except Exception as err:
                raise err
            #finally:
            #    pass

        return True

    def _release_lock(self, obj):
        """Release the lock on a given object.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise ObjectNotInRegistryError"""

        try:
            self.__release_lock(obj)
        except ObjectNotInRegistryError as err:
            pass
        except Exception as err:
            logger.debug("Unknown exception %s" % str(err))
            raise err

    def __release_lock(self, _obj):
        logger.debug("_release_lock")
        obj = stripProxy(_obj)

        if self.find(obj) in self._inprogressDict.keys():
            return

        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot manipulate locks of a disconnected repository!")
        logger.debug("Reg: %s _release_lock(%s)" % (self.name, str(self.find(obj))))
        try:
            if hasattr(obj, '_registry_locked') and obj._registry_locked:
                oid = self.find(obj)
                self.repository.flush([oid])
                obj._registry_locked = False
                self.repository.unlock([oid])
        except (RepositoryError, RegistryAccessError, RegistryLockError, ObjectNotInRegistryError) as err:
            raise err
        except Exception as err:
            logger.debug("un-known registry release lock err!")
            logger.debug("Err: %s" % str(err))
            raise err

    @synchronised
    def pollChangedJobs(self, name):
        """Returns a list of job ids that changed since the last call of this function.
        On first invocation returns a list of all ids.
        "name" should be a unique identifier of the user of this information."""
        logger.debug("pollChangedJobs")
        if self.hasStarted() is True and\
                (time.time() > self._update_index_timer + self.update_index_time):
            changed_ids = self.repository.update_index()
            for this_d in self.changed_ids.itervalues():
                this_d.update(changed_ids)
            self._update_index_timer = time.time()
        res = self.changed_ids.get(name, set(self.ids()))
        self.changed_ids[name] = set()
        return res

    def getIndexCache(self, obj):
        """Returns a dictionary to be put into obj._index_cache through setNodeIndexCache
        This can and should be overwritten by derived Registries to provide more index values."""
        return {}

    @synchronised
    def startup(self):
        """Connect the repository to the registry. Called from Repository_runtime.py"""
        try:
            self._hasStarted = True
            t0 = time.time()
            self.repository = makeRepository(self)
            self._objects = self.repository.objects
            self._incomplete_objects = self.repository.incomplete_objects

            if self._needs_metadata:
                t2 = time.time()
                if self.metadata is None:
                    self.metadata = Registry(self.name + ".metadata", "Metadata repository for %s" % self.name, dirty_flush_counter=self.dirty_flush_counter, update_index_time=self.update_index_time)
                    self.metadata.type = self.type
                    self.metadata.location = self.location
                    setattr(self.metadata, '_parent', self) ## rcurrie Registry has NO '_parent' Object so don't understand this is this used for JobTree?
                logger.debug("metadata startup")
                self.metadata.startup()
                t3 = time.time()
                logger.debug("Startup of %s.metadata took %s sec" % (str(self.name), str(t3-t2)))

            logger.debug("repo startup")
            #self.hasStarted() = True
            self.repository.startup()
            # All Ids could have changed
            self.changed_ids = {}
            t1 = time.time()
            logger.debug("Registry '%s' [%s] startup time: %s sec" % (self.name, self.type, t1 - t0))
        except Exception as err:
            logger.debug("Logging Repo startup Error: %s" % str(err))
            self._hasStarted = False
            raise err

    @synchronised
    def shutdown(self):
        """Flush and disconnect the repository. Called from Repository_runtime.py """
        from Ganga.Utility.logging import getLogger
#        self.shouldReleaseRun = False
#        self.releaseThread.stop()
        logger = getLogger()
        logger.debug("Shutting Down Registry")
        logger.debug("shutdown")
        try:
            self._hasStarted = True
            try:
                if not self.metadata is None:
                    try:
                        self.flush_all()
                    except Exception, err:
                        logger.debug("shutdown _flush Exception: %s" % str(err))
                    self.metadata.shutdown()
            except Exception as err:
                logger.debug("Exception on shutting down metadata repository '%s' registry: %s", self.name, str(err))
            #finally:
            #    pass
            try:
                self.flush_all()
            except Exception as err:
                logger.error("Exception on flushing '%s' registry: %s", self.name, str(err))
                #raise err
            #finally:
            #    pass
            for obj in self._objects.values():
                # locks are not guaranteed to survive repository shutdown
                obj._registry_locked = False
            self.repository.shutdown()

            self._loaded_ids = []

            self.metadata = None

        finally:
            self._hasStarted = False

    def info(self, full=False):
        """Returns an informative string onFlush and disconnect the repository. Called from Repository_runtime.py """
        logger.debug("info")
        s = "registry '%s': %i objects" % (self.name, len(self._objects))
        if full:
            other_sessions = self.repository.get_other_sessions()
            if len(other_sessions) > 0:
                s += ", %i other concurrent sessions:\n * %s" % (len(other_sessions), "\n * ".join(other_sessions))
        return s

    def print_other_sessions(self):
        other_sessions = self.repository.get_other_sessions()
        if len(other_sessions) > 0:
            logger.warning("%i other concurrent sessions:\n * %s" % (len(other_sessions), "\n * ".join(other_sessions)))

    def has_loaded(self, obj):
        """Returns True/False for if a given object has been fully loaded by the Registry.
        Returns False on the object not being in the Registry!
        This ONLY applies to master jobs as the Registry has no apriori knowledge of the subjob structure.
        Consult SubJobXMLList for a more fine grained loaded/not-loaded info/test."""
        try:
            index = self.find(obj)
        except ObjectNotInRegistryError:
            return False

        if index in self._loaded_ids:
            return True
        else:
            return False




from __future__ import absolute_import
from .ArgSplitter import ArgSplitter
#from .ExeSplitter import ExeSplitter
from .GenericSplitter import GenericSplitter
from .GangaDatasetSplitter import GangaDatasetSplitter
#from .DefaultSplitter import DefaultSplitter



from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks import Transform
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import DQ2Dataset, DQ2OutputDataset
from GangaAtlas.Lib.Athena.DQ2JobSplitter import DQ2JobSplitter
from TaskApplication import AthenaTask, AnaTaskSplitterJob
from Ganga.GPIDev.Schema import *

from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException, DQDatasetExistsException, DQFileExistsInDatasetException, DQInvalidRequestException
from dq2.container.exceptions import DQContainerAlreadyHasDataset
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_lock, dq2

import time

from dq2.info import TiersOfATLAS
from GangaAtlas.Lib.ATLASDataset import whichCloud
from Ganga.Core.exceptions import ApplicationConfigurationError

from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname 

PandaClient = None
def getPandaClient():
    global PandaClient
    if PandaClient is None:
        from pandatools import Client
        PandaClient = Client
    return PandaClient

import random

def whichCloudExt(site):
   if site.startswith("NDGF"):
      return "NG"
   return whichCloud(site)

def stripSite(site):
   dq2alternatename = TiersOfATLAS.getSiteProperty(site,'alternateName')
   if not dq2alternatename:
      return site
   else: 
      return dq2alternatename[0]

def stripSites(sites):
   newsites = {}
   for site in sites:
      newsites[stripSite(site)] = 1
   return newsites.keys()


class AnaTransform(Transform):
   """ Analyzes Events """
   _schema = Schema(Version(1,0), dict(Transform._schema.datadict.items() + {
       'files_per_job'     : SimpleItem(defvalue=5, doc='files per job', modelist=["int"]),
       'partitions_data'   : ComponentItem('datasets', defvalue=[], optional=1, sequence=1, hidden=1, doc='Input dataset for each partition'),
       'partitions_sites'  : SimpleItem(defvalue=[], hidden=1, modelist=["str","list"],doc='Input site for each partition'),
       'outputdata'        : ComponentItem('datasets', defvalue=DQ2OutputDataset(), doc='Output dataset'),
       'dataset_name'      : SimpleItem(defvalue="", transient=1, comparable=False, getter="get_dataset_name", doc='name of the output dataset'),
       }.items()))
   _category = 'transforms'
   _name = 'AnaTransform'
   _exportmethods = Transform._exportmethods

   def initialize(self):
      super(AnaTransform, self).initialize()
      self.application = AthenaTask()
      self.inputdata = DQ2Dataset()

   def get_dataset_name(self):
      task = self._getParent()
      name_base = ["user",getNickname(),task.creation_date,"task_%s" % task.id]
      if self.inputdata.dataset:
          subtask_dsname = ".".join(name_base +["subtask_%s" % task.transforms.index(self), str(self.inputdata.dataset[0].strip("/"))])
      else:
          subtask_dsname = ".".join(name_base +["subtask_%s" % task.transforms.index(self)])
            
      # make sure we keep the name size limit:
      dq2_config = getConfig("DQ2")
      max_length_site = len("ALBERTA-WESTGRID-T2_SCRATCHDISK      ")
      max_length = dq2_config['OUTPUTDATASET_NAMELENGTH'] - max_length_site
      if len(subtask_dsname) > max_length:
          logger.debug("Proposed dataset name longer than limit (%d). Restricting dataset name..." % max_length)

          while len(subtask_dsname) > max_length:
              subtask_dsname_toks = subtask_dsname.split('.')
              subtask_dsname = '.'.join(subtask_dsname_toks[:len(subtask_dsname_toks)-1])
      return subtask_dsname

## Internal methods
   def checkCompletedApp(self, app):
      task = self._getParent()
      j = app._getParent()
      for odat in j.outputdata.outputdata:
          # Look out: if this is changed, there is anothher one like it below!
          if 0==len([f for f in j.outputdata.output if ".".join(odat.split(".")[:-1]) in f]):
              logger.error("Job %s has not produced %s file, only: %s" % (j.id, odat, j.outputdata.output))
              return False
      # if this is the first app to complete the partition...
      if self.getPartitionStatus(self._app_partition[app.id]) != "completed":
          task_container, subtask_dsname = task.container_name, self.dataset_name

          infos = {}
          for oinfo in j.outputdata.output:
              try:
                  dq2_lock.acquire()
                  info = oinfo.split(",")
                  # get master replica from dataset - info not set to SE; but to ANALY_XYZ from panda
                  master_replica = dq2.getMasterReplicaLocation(info[0])
                  if master_replica:
                      info[5] = master_replica
                  else:
                      replicas = dq2.listDatasetReplicas(info[0]).values()
                      if len(replicas) == 0:
                          try:
                              info[5] = getPandaClient().PandaSites[info[5]]["ddm"]
                          except KeyError:
                              pass
                      else:
                          complete, incomplete = replicas[0].values()
                          info[5] = (complete + incomplete)[0]
                  if info[4][:3] == "ad:":
                      info[4] = info[4][3:]

              finally:
                  dq2_lock.release()
                
              datasetname = subtask_dsname + '.' + info[5]
              info[0] = datasetname
              infos.setdefault(datasetname, []).append(",".join(info))

          for ds in infos.keys():
              outputdata = DQ2OutputDataset()
              try:
                  outputdata.create_dataset(ds)
              except DQDatasetExistsException:
                  pass
              try:
                  outputdata.register_datasets_details(None, infos[ds])
              except DQFileExistsInDatasetException:
                  pass

          # Register Container
          try:
              containerinfo = {}
              dq2_lock.acquire()
              try:
                  containerinfo = dq2.listDatasets(task_container)
              except:
                  containerinfo = {}
              if containerinfo == {}:
                  try:
                      dq2.registerContainer(task_container)
                      logger.debug('Registered container for Task %i: %s' % (task.id, task_container))
                  except Exception as x:
                      logger.error('Problem registering container for Task %i, %s : %s %s' % (task.id, task_container,x.__class__, x))
              for ds in infos.keys():
                  try:
                      dq2.registerDatasetsInContainer(task_container, [ ds ] )
                  except DQContainerAlreadyHasDataset:
                      pass
                  except Exception as x:
                      logger.error('Problem registering dataset %s in container %s: %s %s' %( subtask_dsname, task_container, x.__class__, x))
          finally:
              dq2_lock.release()
      return True

   def rebrokerPanda(self, cloud = None):
      sites = getPandaClient().PandaSites.keys()

      dataset_sites = [getPandaClient().convertDQ2toPandaID(site) for site in self.inputdata.get_locations(complete=1)]
      sites = [site for site in sites if site in dataset_sites]
      sites = [s for s in sites if not s in self.backend.requirements.excluded_sites] 
      sites = [s for s in sites if not s.replace("ANALY_","") in self.backend.requirements.excluded_sites]
      if cloud:
         sites = [s for s in sites if s.replace("ANALY_","") in getPandaClient().PandaClouds[cloud]["sites"]] 
      if len(sites) == 0:
         logger.error("No compatible sites for rebrokering found!")
         return

      from random import shuffle
      shuffle(sites)
      self.backend.site = sites[0]
      self.partitions_sites = [sites[0]]*len(self.partitions_sites)
      logger.warning("Rebrokering transform - %i possible sites; chosing %s at random..." % (len(sites),sites[0]))


   def findCompleteCloudBackend(self,db_sites,allowed_sites,replicas):
      # Sort complete replicas into clouds
      # returns sorted list of tuples (cloud, backend)
      addclouds = TiersOfATLAS.ToACache.dbcloud.keys()
      random.shuffle(addclouds)
      clouds = config["cloudPreference"] + addclouds
      complete_sites = {}
      for c in clouds:
         complete_sites[c] = []
      for s in replicas[1]:
         try:
            cloud = whichCloudExt(s)
         except:
            logger.warning("Could not get cloud of site %s!", s)
            continue
         if db_sites and not stripSite(s) in db_sites:
            continue
         #print cloud, s
         complete_sites[cloud].append(s)

      result = []
      # Check if we find a cloud/backend combination for a complete replica
      backends = [be for be in config["backendPreference"] if be in GPI.__dict__]
      for cloud in clouds:
         for backend in backends:
            if not backend in allowed_sites.keys(): 
               continue
            if backend == "Panda":
                sites = [site for site in complete_sites[cloud] if getPandaClient().convertDQ2toPandaID(site) in allowed_sites["Panda"]]
            else:
                sites = [site for site in complete_sites[cloud] if site in allowed_sites[backend]]
            if len(sites) > 0:
               result.append((cloud,backend))
      return result
 
   def findIncompleteCloudBackend(self,db_sites,allowed_sites,replicas):
      # If no cloud/backend combination is found for complete replicas, 
      # find cloud/backend with maximal number of replicas
      # returns list of tuples sorted by number of sites with replicas: (cloud, backend)
      addclouds = TiersOfATLAS.ToACache.dbcloud.keys()
      random.shuffle(addclouds)
      clouds = config["cloudPreference"] + addclouds
      incomplete_sites = {}
      for c in clouds:
         incomplete_sites[c] = []
      for s in replicas[0]:
         try:
            cloud = whichCloudExt(s)
         except:
            continue
         if db_sites and not stripSite(s) in db_sites:
            continue
         incomplete_sites[cloud].append(s)
      
      using_backend = None
      using_cloud = None
      max_sites = 0
      result = []
      backends = [be for be in config["backendPreference"] if be in GPI.__dict__]
      for cloud in clouds:
         for backend in backends:
            if backend == "Panda":
                from pandatools import Client
                sites = [site for site in incomplete_sites[cloud] if getPandaClient().convertDQ2toPandaID(site) in allowed_sites["Panda"]]
            else:
                sites = [site for site in incomplete_sites[cloud] if site in allowed_sites[backend]]
            if len(sites) > 0:
               result.append((len(sites),cloud,backend))
      result.sort()
      return [(r[1],r[2]) for r in result]

   def check(self):
      super(AnaTransform,self).check()
      if not self.inputdata.dataset:
         return
      if not self.backend:
         logger.warning("Determining backend and cloud...")

         # Get ddm sites of atlas_dbrelease, if present
         db_sites = None
         if self.application.atlas_dbrelease == "LATEST":
            from pandatools import Client
            self.application.atlas_dbrelease = getPandaClient().getLatestDBRelease(False)
         if self.application.atlas_dbrelease:
            try:
               db_dataset = self.application.atlas_dbrelease.split(':')[0] 
               try:
                  dq2_lock.acquire()
                  db_locations = dq2.listDatasetReplicas(db_dataset).values()[0][1] 
               finally:
                  dq2_lock.release()

            except Exception as x:
               raise ApplicationConfigurationError(x, 'Problem in AnaTask - j.application.atlas_dbrelease is wrongly configured ! ')
            db_sites = stripSites(db_locations)
         
         # Get complete/incomplete ddm sites for input dataset
         ds = self.inputdata.dataset[0]
         try:
            dq2_lock.acquire()
            if ds[-1] != "/":
               try:
                  replicas = {ds : dq2.listDatasetReplicas(ds)}
               except DQUnknownDatasetException:
                  ds += "/"
            if ds[-1] == "/":
               replicas = dq2.listDatasetReplicasInContainer(ds)
         finally:
            dq2_lock.release()
        
         # check if replicas are non-empty
         somefound = False
         for tid in replicas: 
            if len(replicas[tid]) == 0:
               raise ApplicationConfigurationError(None, "No replicas for dataset %s found!" % tid)
         replicas = [r.values()[0] for r in replicas.values()] # (dict with only one entry)

         # Get allowed sites for each backend:
         backends = [be for be in config["backendPreference"] if be in GPI.__dict__]
         allowed_sites = {}
         if "LCG" in backends:
            allowed_sites["LCG"] = GPI.LCG().requirements.list_sites(True,True)
         if "Panda" in backends:
            from pandatools import Client
            allowed_sites["Panda"] = getPandaClient().PandaSites.keys()
            #allowed_sites["Panda"] = [site["ddm"] for site in Client.getSiteSpecs()[1].values()]
         if "NG" in backends:
            allowed_sites["NG"] = getConfig("Athena")["AllowedSitesNGDQ2JobSplitter"]
         #if "PBS" in backends:
         #   sites["PBS"] = [] # should be local DQ2 storage element!

         # Get list of cloud-backend pairs (cbl) for complete replicas
         common_cbl = None
         for r in replicas:
            cbl = self.findCompleteCloudBackend(db_sites, allowed_sites, r)
            if common_cbl is None:
               common_cbl = cbl
            else:
               common_cbl = [cb for cb in cbl if cb in common_cbl]

         #print "CLOUD/BACKEND list for COMPLETE replicas: ", common_cbl

         # ..and for incomplete replicas
         if common_cbl is None or len(common_cbl) == 0:
            if len(replicas) > 1:
               raise ApplicationConfigurationError(None, 'Container dataset %s has no complete replica on one site and backend. Please specify individual tid datasets or use t.initializeFromDataset("%s") ' % (ds, ds))
            common_cbl = self.findIncompleteCloudBackend(db_sites, allowed_sites, replicas[0])
            #print "CLOUD/BACKEND list for INCOMPLETE replicas: ", common_cbl
         if common_cbl is None or len(common_cbl) == 0:
            raise ApplicationConfigurationError(None, 'Container dataset %s has no replica on one site and backend. Please specify individual tid datasets!' % (ds))

         cb = common_cbl[0]
         using_cloud = cb[0]
         using_backend = cb[1]

         assert using_cloud, using_backend

         if using_backend == "Panda":
            self.backend = stripProxy(GPI.Panda())
            self.backend.requirements.cloud = using_cloud

         elif using_backend == "NG":
            self.backend = stripProxy(GPI.NG())
         elif using_backend == "LCG":
            self.backend = stripProxy(GPI.LCG())
            self.backend.requirements = stripProxy(GPI.AtlasLCGRequirements())
            self.backend.requirements.cloud = using_cloud
         assert self.backend
         logger.warning("Running on cloud %s using backend %s", using_cloud, using_backend)

      logger.warning("Determining partition splitting...")
      try:
         if not self.backend.requirements.cloud:
            self.backend.requirements.cloud = "DE"
      except:
         pass
      if not self.inputdata.dataset:
         return
      splitter = DQ2JobSplitter()
      splitter.numfiles = self.files_per_job
      #splitter.update_siteindex = False # commented to use default value
      #splitter.use_lfc = True
      sjl = splitter.split(self) # This works even for Panda, no special "Job" properties are used anywhere.
      self.partitions_data = [sj.inputdata for sj in sjl]
      try:
         self.partitions_sites = [sj.backend.requirements.sites for sj in sjl]
      except AttributeError:
         self.partitions_sites = [sj.backend.site for sj in sjl]
         pass
      self.setPartitionsLimit(len(self.partitions_data)+1)
      self.setPartitionsStatus([c for c in range(1,len(self.partitions_data)+1) if self.getPartitionStatus(c) != "completed"], "ready")
   
   def getJobsForPartitions(self, partitions):
      j = self.createNewJob(partitions[0])
      if len(partitions) >= 1:
          j.splitter = AnaTaskSplitterJob()
          j.splitter.subjobs = partitions
      j.inputdata = self.partitions_data[partitions[0]-1]
      if self.partitions_sites:
         if stripProxy(j.backend)._name == "Panda":
            if j.backend.site == "AUTO":
               j.backend.site = self.partitions_sites[partitions[0]-1]
         else:
            j.backend.requirements.sites = self.partitions_sites[partitions[0]-1]
      j.outputdata = self.outputdata
      if stripProxy(j.backend)._name == "Panda" and j.application.atlas_exetype == "ATHENA":
          j.outputdata.outputdata=[]
      #j.outputdata.datasetname = ""
      task = self._getParent()
      dsn = ["user",getNickname(),task.creation_date,"%i.t_%s_%s" % (j.id, task.id, task.transforms.index(self))]
      j.outputdata.datasetname = ".".join(dsn)

      #if j.outputdata.datasetname:
         #today = time.strftime("%Y%m%d",time.localtime())
         #j.outputdata.datasetname = "%s.%i.%s" % (j.outputdata.datasetname, j.id, today)
      return [j]

   def info(self):
      print markup("%s '%s'" % (self.__class__.__name__, self.name), status_colours[self.status])
      print "* dataset: %s " % self.inputdata.dataset
      print "* processing %s per job" % say(self.files_per_job,"file")
      print "* backend: %s" % self.backend.__class__.__name__
      print "* application:"
      self.application.printTree() 





#!/usr/bin/env python

import os, re, pickle

########################################################################
if __name__ == '__main__':

    stats = {}

    print '-------------------'
    print 'getstats called ...'

    # Determine dataset access mode
    try:
        datasettype = os.environ['DATASETTYPE']
    except:
        print "ERROR : DATASETTYPE not defined, using DQ2_LOCAL"
        datasettype = 'DQ2_LOCAL'
        pass
    # use DQ2_LOCAL as default
    if not datasettype in [ 'DQ2_LOCAL', 'DQ2_COPY', 'FILE_STAGER' ]:
        datasettype = 'DQ2_LOCAL'

    numfiles3 = 0
    filepat = '"PFN:(.+)"'
    selectorpat = 'ServiceMgr.EventSelector.InputCollections = \[(.+),\]'
    selectorpat2 = 'mySampleList = \[(.+)\]'
    sitepat = 'DQ2_LOCAL_SITE_ID: (.+)'
    datasetnamepat = '^DATASETNAME=(.+)'
    
    input_files = []
    input_files_sel = []

    # collect stats from stderr
    percentcpu = 0
    ipercentcpu = 0
    wallclock = 0
    usertime = 0
    systemtime = 0
    try:
        if 'stderr' in os.listdir('.'):  
            zfile = os.popen('cat '+os.path.join('.','stderr' ))
            for line in zfile:
                
                if line.find('Percent of CPU this job got')>-1:
                    try:
                        percentcpu = percentcpu + int(re.match('.*got: (.*).',line).group(1))
                    except ValueError:
                        percentcpu = 0  
                    ipercentcpu = ipercentcpu + 1
                if line.find('Elapsed (wall clock) time')>-1:
                    try:
                        iwallclock = re.match('.*m:ss\): (.*)\.\d\d',line).group(1).split(':')
                        wallclock = wallclock + int(iwallclock[0])*60+int(iwallclock[1])
                    except:
                        iwallclock = re.match('.*m:ss\): (.*)',line).group(1).split(':')
                        wallclock = wallclock + int(iwallclock[0])*3600+int(iwallclock[1])*60+int(iwallclock[2])
                if line.find('User time (seconds)')>-1:
                    iusertime = float(re.match('.*User time \(seconds\): (.*)',line).group(1))
                    usertime = usertime + iusertime
                if line.find('System time (seconds)')>-1:
                    isystemtime = float(re.match('.*System time \(seconds\): (.*)',line).group(1))
                    systemtime = systemtime + isystemtime
                if line.find('Exit status')>-1:
                    stats['exitstatus'] = re.match('.*status: (.*)',line).group(1)
                if line.find('can not be opened for reading (Timed out)')>-1:
                    stats['filetimedout'] = True

            if ipercentcpu > 0:            
                stats['percentcpu'] = percentcpu / ipercentcpu
                stats['usertime'] = usertime
                stats['systemtime'] = systemtime
                stats['wallclock'] = wallclock
            else:
                stats['percentcpu'] = 0
                stats['wallclock'] = 0
                stats['usertime'] = 0
                stats['systemtime'] = 0
            if zfile:        
                zfile.close()

    except MemoryError:
        print 'ERROR stderr logfiles too large to be parsed.'
        pass
    
    # collect stats from stdout
    totalevents = 0
    itotalevents = 0
    jtotalevents = 0
    numfiles = 0
    numfiles2 = 0
    numfiles3 = 0
    try:
        if 'stdout' in os.listdir('.'):  
            zfile = os.popen('cat '+os.path.join('.','stdout' ))
            for line in zfile:
                if line.find('Storing file at:')>-1:
                    stats['outse'] = re.match('.*at: (.*)',line).group(1)
                if line.find('SITE_NAME=')>-1:
                    stats['site'] = re.match('SITE_NAME=(.*)',line).group(1)
                #if line.find('Database being retired...')>-1:
                #    stats['dbretired'] = True
                if line.find('Core dump from CoreDumpSvc')>-1:
                    stats['coredump'] = True
                if line.find('Cannot load entry')>-1:
                    stats['cannotloadentry'] = True
                if line.find('cannot open a ROOT file in mode READ if it does not exists')>-1:
                    stats['filenotexist'] = True
                if line.find('FATAL finalize: Invalid state "Configured"')>-1:
                    stats['invalidstateconfig'] = True
                if line.find('failure in an algorithm execute')>-1:
                    stats['failalg'] = True
                if line.find('events processed so far')>-1:
                    try:
                        itotalevents = int(re.match('.* run #\d+ (\d+) events processed so far.*',line).group(1))
                        jtotalevents = int(itotalevents)
                    except:
                        pass
                if line.find('cObj_DataHeader...')>-1:
                    numfiles2 = numfiles2 + int(re.match('.* #=(.*)',line).group(1))
                if line.find('"PFN:')>-1:
                    numfiles3 = numfiles3 + 1
                if line.find('rfio://')>-1 and line.find('Always Root file version')>-1:
                    try:
                        stats['server'] = re.match('(.+://.+)//.*',line).group(1)
                    except:
                        stats['server'] = 'unknown'

                if line.find('Info Database being retired...')>-1:
                    numfiles = numfiles + 1
                    totalevents = totalevents + itotalevents
                    itotalevents = 0
                if line.find('GANGATIME1')==0:
                    stats['gangatime1'] = int(re.match('GANGATIME1=(.*)',line).group(1))
                if line.find('GANGATIME2')==0:
                    stats['gangatime2'] = int(re.match('GANGATIME2=(.*)',line).group(1))
                if line.find('GANGATIME3')==0:
                    stats['gangatime3'] = int(re.match('GANGATIME3=(.*)',line).group(1))
                if line.find('GANGATIME4')==0:
                    stats['gangatime4'] = int(re.match('GANGATIME4=(.*)',line).group(1))
                if line.find('GANGATIME5')==0:
                    stats['gangatime5'] = int(re.match('GANGATIME5=(.*)',line).group(1))

                try:
                    if line.find('NET_ETH_RX_PREATHENA')==0:
                        stats['NET_ETH_RX_PREATHENA'] = int(re.match('NET_ETH_RX_PREATHENA=(.*)',line).group(1))
                    if line.find('NET_ETH_RX_AFTERATHENA')==0:
                        stats['NET_ETH_RX_AFTERATHENA'] = int(re.match('NET_ETH_RX_AFTERATHENA=(.*)',line).group(1))
                except:
                    stats['NET_ETH_RX_PREATHENA'] = 0
                    stats['NET_ETH_RX_AFTERATHENA'] = 0

                try:
                    if line.find('### node info:')==0:
                        stats['arch'] = re.match('### node info: .*,(.*, .*),.*,.*,.*,.*', line).group(1).strip()
                except:
                    stats['arch'] = ''

            stats['numfiles2'] = numfiles2

            if datasettype == 'DQ2_COPY':
                stats['numfiles'] = numfiles / 2
                stats['totalevents'] = totalevents
                stats['numfiles3'] = numfiles3 / 2
            elif datasettype == 'FILE_STAGER':
                stats['numfiles'] = (numfiles - 2)/2
                stats['totalevents'] = jtotalevents
                stats['numfiles3'] = numfiles3 - 1
            else:
                stats['numfiles'] = numfiles - 1
                stats['totalevents'] = jtotalevents
                stats['numfiles3'] = numfiles3 - 1

            if zfile:        
                zfile.close()

    except MemoryError:
        print 'ERROR stderr logfile too large to be parsed.'
        pass

    # collect stats from AthSummary.txt
    try:
        if 'AthSummary.txt' in os.listdir('.'):  
            zfile = os.popen('cat '+os.path.join('.','AthSummary.txt' ))
            for line in zfile:
                if line.find('Files read:')==0:
                    stats['numfiles'] = int(re.match('Files read:(.*)',line).group(1))
                    stats['numfiles2'] = stats['numfiles'] 
                    stats['numfiles3'] = stats['numfiles'] 
                if line.find('Events Read:')==0:
                    stats['totalevents'] = int(re.match('Events Read:(.*)',line).group(1))

            if zfile:        
                zfile.close()

    except MemoryError:
        print 'ERROR AthSummary.txt logfile too large to be parsed.'
        pass

    print stats

    f = open('stats.pickle','w')
    pickle.dump(stats,f)
    f.close()

    print 'getstats finished...' 
    print '-------------------'




from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaAtlas.Lib.Credentials.ProxyHelper import getNickname 
import time
from Ganga.GPIDev.Lib.Tasks import ITask
from GangaAtlas.Lib.Tasks.AtlasTransform import AtlasTransform
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import dq2_lock, dq2
from dq2.common.DQException import DQException
from dq2.clientapi.DQ2 import DQ2, DQUnknownDatasetException, DQDatasetExistsException, DQFileExistsInDatasetException, DQInvalidRequestException
from dq2.container.exceptions import DQContainerAlreadyHasDataset, DQContainerDoesNotHaveDataset
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import getConfig
configDQ2 = getConfig('DQ2')

########################################################################

class AtlasTask(ITask):
    """Atlas add-ons for the Task framework"""
    _schema = Schema(Version(1,0), dict(ITask._schema.datadict.items() + {
        }.items()))
    
    _category = 'tasks'
    _name = 'AtlasTask'
    _exportmethods = ITask._exportmethods + [ 'getContainerName', 'initializeFromDictionary', 'checkOutputContainers', 'checkInputDatasets' ]

    _tasktype = "ITask"
    
    default_registry = "tasks"
    
    def getContainerName(self, max_length = configDQ2['OUTPUTDATASET_NAMELENGTH'] - 2):
        if self.name == "":
            name = "task"
        else:
            name = self.name

        # check container name isn't too big            
        dsn = ["user",getNickname(),self.creation_date, name, "id_%i/" % self.id ]
        if len(".".join(dsn)) > max_length:
            dsn = ["user",getNickname(),self.creation_date, name[: - (len(".".join(dsn)) - max_length)], "id_%i/" % self.id ]

        return (".".join(dsn)).replace(":", "_").replace(" ", "").replace(",","_")

    def checkOutputContainers(self):
        """Go through all transforms and check all datasets are registered"""
        logger.info("Cleaning out overall Task container...")

        try:
            dslist = []
            dq2_lock.acquire()
            try:
                dslist = dq2.listDatasetsInContainer(self.getContainerName())
            except:
                dslist = []

            try:
                dq2.deleteDatasetsFromContainer(self.getContainerName(), dslist )

            except DQContainerDoesNotHaveDataset:
                  pass
            except Exception as x:
                logger.error("Problem cleaning out Task container: %s %s", x.__class__, x)
            except DQException as x:
                logger.error('DQ2 Problem cleaning out Task container: %s %s' %( x.__class__, x))
        finally:
            dq2_lock.release()

        logger.info("Checking output data has been registered. This can take a few minutes...")
        for trf in self.transforms:
            logger.info("Checking containers in Tranform %d..." % trf.getID() )
            trf.checkOutputContainers()

    def initializeFromDictionary(self, ds_dict, template = None, files_per_job = -1, MB_per_job = -1, subjobs_per_unit = -1,
                                 application = None, backend = None):
        """Initialize transforms from a dictionary of containers"""
        if self.status != "new":
            logger.error("Cannot add more data to a new task yet. Give me time :)")
            return
        
        for trf_name, physics_container in ds_dict.iteritems():
            trf = AtlasTransform()
            trf.name = trf_name
            self.appendTransform(trf)
            if files_per_job > 0:
                trf.files_per_job = files_per_job
            if MB_per_job > 0:
                trf.MB_per_job = MB_per_job
            if subjobs_per_unit > 0:
                trf.subjobs_per_unit = subjobs_per_unit
            trf.application = application._impl.clone()
            trf.backend = backend._impl.clone()
            trf.initializeFromContainer(physics_container, template)

    def checkInputDatasets(self):
        """Check the distribution of the input datasets"""
        for trf in self.transforms:
            trf.checkInputDatasets()



from __future__ import print_function
from __future__ import absolute_import

import Ganga.Utility.logging as logging


def test_logging():
    logger = logging.getLogger()

    logger.info('info test')
    logger.error('error test')
    logger.warning('warning test')

    private_logger = logging._get_logging().getLogger("TESTLOGGER.CHILD.GRANDCHILD")
    formatter = logging._get_logging().Formatter(logging._formats['DEBUG'])
    console = logging._get_logging().StreamHandler()
    console.setFormatter(formatter)
    private_logger.setLevel(logging._get_logging().DEBUG)
    private_logger.addHandler(console)
    private_logger.critical('hello')



"""Object model for extract data as handled by base action implementations.

The Node class represents a node of extracted data containing a value and / or
subnodes.

"""

from xml.dom.minidom import getDOMImplementation, parseString


class Node(object):
    
    """A node of extracted data containing a value and / or subnodes.
    
    Methods are provided for creating a subtree of nodes, see addnode(), and for
    querying that subtree, see getnode(), getnodes(), getvalue() and
    getvalues().
    
    The value and subnodes can also be accessed directly via the value and nodes
    attributes, respectively.

    In addition methods are provided to convert to and from XML DOMs and
    strings, see todom(), fromdom(), toxml(), fromxml() and toprettyxml().
    
    For simplicity, Node supports only a subset of XML; essentially it does not
    support attributes.
    
    Example representation of Node in XML (referenced in method documentation):
    <run>
        <core>
            <id>2007-06-22_13.17.51</id>
            <start-time>2007/06/22 13:17:51</start-time>
            <extract-time>2007/06/22 13:18:55</extract-time>
        </core>
        <job>
            <core>
                <id>4</id>
                <name>Core_2007-06-22_13.17.51_0</name>
                <status>completed</status>
                <application>Executable</application>
                <backend>Local</backend>
                <backend-id>19253</backend-id>
                <backend-actualCE>lx09.hep.ph.ic.ac.uk</backend-actualCE>
            </core>
        </job>
        <job>
            <core>
                <id>5</id>
                <name>Core_2007-06-22_13.17.51_1</name>
                <status>completed</status>
                <application>Executable</application>
                <backend>Local</backend>
                <backend-id>19267</backend-id>
                <backend-actualCE>lx09.hep.ph.ic.ac.uk</backend-actualCE>
            </core>
        </job>
    </run>
    
    """
    
    def __init__(self, name, value = None):
        """Create a new node with the given name and value.

        Keyword arguments:
        name -- The node name.
        value -- Optional value, which is converted to a string using str() if
            not None.

        """
        self.name = name
        self.value = None
        if value != None:
            self.value = str(value)
        self.nodes = []

    def addnode(self, name, value = None):
        """Add a new node with the given name and value, returning the new node.
        
        Keyword arguments:
        name -- The node name.
        value -- Optional value, which is converted to a string using str() if
            not None.
        
        Example - creating a node tree (using example in class documentation):

        run = Node('run')
        run_core = run.addnode('core')
        run_core.addnode('id', '2007-06-22_13.17.51')
        run_core.addnode('start-time', '2007/06/22 13:17:51')
        run_core.addnode('extract-time', '2007/06/22 13:18:55')
        job = run.addnode('job')
        job_core = job.addnode('core')
        job_core.addnode('id', 4)
        job_core.addnode('name', 'Core_2007-06-22_13.17.51_0')
        job_core.addnode('status', 'completed')
        ...

        """
        subnode = Node(name, value)
        self.nodes.append(subnode)
        return subnode
    
    def getnode(self, path = '', sep = '.', wild='*'):
        """Return the first node matching the given path, or None if no matches.

        Keyword arguments:
        path -- Optional path to node.
        sep -- Optional path separator, normally left as default.
        wild -- Optional wildcard, normally left as default.

        This method effectively returns the first element of the list returned
        by getnodes(), or None if the list is empty. See getnodes() for path
        usage and examples.

        """
        selected = self.getnodes(path, 1, sep, wild)
        if selected:
            return selected[0]
        else:
            return None

    def getnodes(self, path = '', maxresults = -1, sep = '.', wild='*'):
        """Return a list of all nodes matching the given path, or an empty list if no matches.
        
        Keyword arguments:
        path -- Optional path to node.
        maxresults -- Optional maximum number of results, non-positive indicates
            no limit.
        sep -- Optional path separator, normally left as default.
        wild -- Optional wildcard, normally left as default.
        
        The path is, by default, '.' separated with '*' wildcard. The returned
        list is in recursive order of the subnodes; e.g. the path 'job.core'
        returns a list of all 'core' subnodes of the first 'job' subnode
        followed by all 'core' subnodes of the second 'job' subnode, etc. See
        examples below for more details.
        
        The returned list is newly created and so modifications to the list
        itself do not modify the structure of the tree. However, the elements of
        the list reference the nodes of the tree and so modifications to the
        elements of the list do modify the tree.
        
        Examples - querying a node tree (using example in class documentation):

        #return list of current node only, (not commonly used).
        run.getnodes() or run.getnodes('')
        
        #return list of direct 'job' subnodes.
        run.getnodes('job')
        
        #return list of all direct subnodes.
        run.getnodes('*')
        
        #return list of 'core' subnodes of 'job' subnodes.
        #i.e. effectively concatenates results of calling getnodes('core') on
        #each node returned by run.getnodes('job').
        run.getnodes('job.core') 

        #return list of 'id' subnodes of 'core' subnodes of 'job' subnodes.
        #i.e. effectively concatenates results of calling getnodes('id') on each
        #node returned by run.getnodes('job.core')
        run.getnodes('job.core.id') 

        #return list of all subnodes of 'job' subnodes.
        #i.e. effectively concatenates results of calling getnodes('*') on each
        #node returned by run.getnodes('job') 
        run.getnodes('job.*')
        
        #return list of all 'id' subnodes of all subnodes of 'job' subnodes. 
        #i.e. effectively concatenates results of calling getnodes('id') on each
        #node returned by run.getnodes('job.*') 
        run.getnodes('job.*.id')
        
        #return list of all 'job' subnodes where 'core.status' has value
        #'completed'.
        [n for n in runnode.getnodes('job') if n.getvalue('core.status') == 'completed']
        
        N.B. the empty string refers to the current node so the following paths
        are equivalent '', '.', '..', as are 'job', '.job', 'job.', etc.
        
        """
        selected = []
        names = path.split(sep)
        if names:
            if names[0] == '':
                subnodes = [self]
            elif names[0] == wild:
                subnodes = self.nodes[:]
            else:
                subnodes = [node for node in self.nodes if node.name == names[0]]
            if len(names) == 1:
                selected.extend(subnodes)
            else:
                subpath = sep.join(names[1:])
                for subnode in subnodes:
                    if maxresults > 0 and len(selected) >= maxresults:
                        break
                    selected.extend(subnode.getnodes(subpath))
        if maxresults > 0:
            return selected[:maxresults]
        else:
            return selected

    def getvalue(self, path = '', sep = '.', wild='*'):
        """Return the value of the first node matching the given path, or None if no matches.

        Keyword arguments:
        path -- Optional path to node.
        sep -- Optional path separator, normally left as default.
        wild -- Optional wildcard, normally left as default.

        This method effectively returns the value of the first element of the
        list returned by getnodes(), or None if the list is empty. See
        getnodes() for path usage and examples.

        """
        values = self.getvalues(path, 1, sep, wild)
        if values:
            return values[0]
        else:
            return None
    
    def getvalues(self, path = '', maxresults = -1, sep = '.', wild='*'):
        """Return a list of the values of all nodes matching the given path, or an empty list if no matches.

        Keyword arguments:
        path -- Optional path to node.
        maxresults -- Optional maximum number of results, non-positive indicates
            no limit.
        sep -- Optional path separator, normally left as default.
        wild -- Optional wildcard, normally left as default.

        This method effectively returns the values of the nodes in the list
        returned by getnodes(). The returned list can contain None, since the
        value of a node can be None. See getnodes() for path usage and examples.

        """
        return [node.value for node in self.getnodes(path, maxresults, sep, wild)]

    def __repr__(self):
        return '<Node %s:%s>' % (self.name, self.value)

    def todom(self):
        """Return an XML DOM representation of the node."""
        document = getDOMImplementation().createDocument(None, self.name, None)
        Node.__fillelement(document, document.documentElement, self)
        return document

    @staticmethod
    def __fillelement(document, element, node):
        if node.value:
            element.appendChild(document.createTextNode(node.value))
        if node.nodes:
            for subnode in node.nodes:
                subelement = document.createElement(subnode.name)
                element.appendChild(subelement)
                Node.__fillelement(document, subelement, subnode)

    @staticmethod
    def fromdom(dom):
        """Return a Node from the given XML DOM.
        
        Keyword attributes:
        dom -- The XML DOM.
        
        Only supported XML features are retained; i.e. attributes are ignored,
        multiple text nodes are stripped of white space and concatenated.
        
        """
        if dom.nodeType == dom.DOCUMENT_NODE:
            return Node.fromdom(dom.documentElement)
        elif dom.nodeType == dom.ELEMENT_NODE:
            node = Node(dom.nodeName)
            value = ''
            for subdom in dom.childNodes:
                if subdom.nodeType == subdom.TEXT_NODE:
                    value = value + subdom.data.strip()
                else:
                    node.nodes.append(Node.fromdom(subdom))
            if value:
                node.value = value
            return node           
        else:
            raise TypeError
    
    def toxml(self):
        """Return a flat XML string representation of the node."""
        return self.todom().toxml()
        
    @staticmethod
    def fromxml(xml):
        """Return a Node from the given XML string.
        
        Keyword attributes:
        xml -- The XML string.
        
        Only supported XML features are retained; i.e. attributes are ignored,
        text nodes are stripped of white space and concatenated.
        
        """
        dom = parseString(xml)
        return Node.fromdom(dom)
    
    def toprettyxml(self):
        """Return a pretty XML string representation of the node."""
        return self.todom().toprettyxml(indent='   ')




from Ganga.GPIDev.Lib.Tasks.common import *
from Ganga.GPIDev.Lib.Tasks.IUnit import IUnit
from Ganga.GPIDev.Lib.Job.Job import JobError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base.Proxy import addProxy, stripProxy

import os

class ND280Unit_CSVEvtList(IUnit):
   _schema = Schema(Version(1,0), dict(IUnit._schema.datadict.items() + {
      'subpartid' : SimpleItem(defvalue=-1,typelist=['int'],doc='Index of this unit which is important when the original CSV file was split by the transform. Otherwise leave it at -1'),
      'eventswanted' : SimpleItem(defvalue='',typelist=['str','list'],doc='CSV list of run, subrun, and event numbers wanted.'),
    }.items()))

   _category = 'units'
   _name = 'ND280Unit_CSVEvtList'
   _exportmethods = IUnit._exportmethods + [ ]

   def __init__(self):
      super(ND280Unit_CSVEvtList, self).__init__()
      
   def createNewJob(self):
      """Create any jobs required for this unit"""      
      j = GPI.Job()
      j._impl.backend = self._getParent().backend.clone()
      j._impl.application = self._getParent().application.clone()
      j.inputdata = self.inputdata.clone()

      trf = self._getParent()
      task = trf._getParent()

      # copy across the outputfiles
      for f in trf.outputfiles:
         j.outputfiles += [f.clone()]

      j.inputsandbox = trf.inputsandbox

      if type(self.eventswanted) == type(''):
        subLines = self.eventswanted
      else:
        subLines = '\n'.join(self.eventswanted)
      # Base for the naming of each subjob's CSV file
      incsvfile = j._impl.application.csvfile
      tmpname = os.path.basename(incsvfile)
      if len(tmpname.split('.')) > 1:
        patterncsv = '.'.join(tmpname.split('.')[0:-1])+"_sub%d."+ tmpname.split('.')[-1]
      else:
        patterncsv = tmpname+"_sub%d"

      from Ganga.GPIDev.Lib.File import FileBuffer
      thiscsv = patterncsv % self.subpartid

      # Create the CSV file for this Unit
      j._impl.getInputWorkspace().writefile(FileBuffer(thiscsv,subLines),executable=0)
      j._impl.application.csvfile = j._impl.getInputWorkspace().getPath()+thiscsv
      j.inputsandbox.append(j._impl.getInputWorkspace().getPath()+thiscsv)

      # Base for the naming of each subjob's output file
      tmpname = os.path.basename(j._impl.application.outputfile)
      if len(tmpname.split('.')) > 1:
        patternout = '.'.join(tmpname.split('.')[0:-1])+"_sub%d."+ tmpname.split('.')[-1]
      else:
        patternout = tmpname+"_sub%d"
      j._impl.application.outputfile = patternout % self.subpartid

      # Sort out the splitter
      if trf.splitter:
         j.splitter = trf.splitter.clone()
         
      return j

   def checkMajorResubmit(self, job):
      """Check if a failed job shold be 'rebrokered' (new job created) rather than just resubmitted"""
      return False

   def majorResubmit(self, job):
      """Do any bookkeeping before a major resubmit is done"""
      super(ND280Unit_CSVEvtList,self).majorResubmit(job)

   def reset(self):
      """Reset the unit completely"""
      super(ND280Unit_CSVEvtList,self).reset()

   def updateStatus(self, status):
      """Update status hook"""
      super(ND280Unit_CSVEvtList,self).updateStatus(status)

   def checkForSubmission(self):
      """Additional checks for unit submission"""

      # call the base class
      if not super(ND280Unit_CSVEvtList,self).checkForSubmission():
         return False

      return True

   def copyOutput(self):
      """Copy the output data to local storage"""

      job = GPI.jobs(self.active_job_ids[0])
      
      if self.copy_output._name != "TaskLocalCopy" or job.outputdata._impl._name != "DQ2OutputDataset":
         logger.error("Cannot transfer from DS type '%s' to '%s'. Please contact plugin developer." % (job.outputdata._name, self.copy_output._name))
         return False

      # check which fies still need downloading
      to_download = []
      for f in job.outputfiles:
         
         # check for REs
         if self.copy_output.isValid( os.path.join(f.localDir, f.namePattern)) and not self.copy_output.isDownloaded( os.path.join(f.localDir, f.namePattern)):
            to_download.append(f)

      # is everything downloaded?
      if len(to_download) == 0:
         return True

      # nope, so pick the requested number and off we go
      for f in to_download:
         f.get()

      return False



from __future__ import absolute_import
from __future__ import print_function
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem, FileItem, GangaFileItem
from .common import logger
from Ganga.Utility.ColourText import status_colours, overview_colours, ANSIMarkup
markup = ANSIMarkup()
import Ganga.GPI as GPI
from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base.Proxy import stripProxy
import time
from Ganga.GPIDev.Lib.Tasks.ITask import addInfoString
import sys
import traceback

def formatTraceback():
   "Helper function to printout a traceback as a string"
   return "\n %s\n%s\n%s\n" % (''.join( traceback.format_tb(sys.exc_info()[2])), sys.exc_info()[0], sys.exc_info()[1])

class IUnit(GangaObject):
    _schema = Schema(Version(1, 0), {
        'status': SimpleItem(defvalue='new', protected=1, copyable=0, doc='Status - running, pause or completed', typelist=["str"]),
        'name': SimpleItem(defvalue='Simple Unit', doc='Name of the unit (cosmetic)', typelist=["str"]),
        'application': ComponentItem('applications', defvalue=None, optional=1, load_default=False, doc='Application of the Transform.'),
        'inputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Input dataset'),
        'outputdata': ComponentItem('datasets', defvalue=None, optional=1, load_default=False, doc='Output dataset'),
        'active': SimpleItem(defvalue=False, hidden=1, doc='Is this unit active'),
        'active_job_ids': SimpleItem(defvalue=[], typelist=['int'], sequence=1, hidden=1, doc='Active job ids associated with this unit'),
        'prev_job_ids': SimpleItem(defvalue=[], typelist=['int'], sequence=1,  hidden=1, doc='Previous job ids associated with this unit'),
        'minor_resub_count': SimpleItem(defvalue=0, hidden=1, doc='Number of minor resubmits'),
        'major_resub_count': SimpleItem(defvalue=0, hidden=1, doc='Number of major resubmits'),
        'req_units': SimpleItem(defvalue=[], typelist=['str'], sequence=1, hidden=1, doc='List of units that must complete for this to start (format TRF_ID:UNIT_ID)'),
        'start_time': SimpleItem(defvalue=0, hidden=1, doc='Start time for this unit. Allows a delay to be put in'),
        'copy_output': ComponentItem('datasets', defvalue=None, load_default=0, optional=1, doc='The dataset to copy the output of this unit to, e.g. Grid dataset -> Local Dataset'),
        'merger': ComponentItem('mergers', defvalue=None, load_default=0, optional=1, doc='Merger to be run after this unit completes.'),
        'splitter': ComponentItem('splitters', defvalue=None, optional=1, load_default=False, doc='Splitter used on each unit of the Transform.'),
        'postprocessors': ComponentItem('postprocessor', defvalue=None, doc='list of postprocessors to run after job has finished'),
        'inputsandbox': FileItem(defvalue=[], typelist=['str', 'Ganga.GPIDev.Lib.File.File.File'], sequence=1, doc="list of File objects shipped to the worker node "),
        'inputfiles': GangaFileItem(defvalue=[], typelist=['str', 'Ganga.GPIDev.Adapters.IGangaFile.IGangaFile'], sequence=1, doc="list of file objects that will act as input files for a job"),
        'outputfiles': GangaFileItem(defvalue=[], typelist=['str', 'Ganga.GPIDev.Adapters.IGangaFile.IGangaFile'], sequence=1, doc="list of OutputFile objects to be copied to all jobs"),
        'info' : SimpleItem(defvalue=[],typelist=['str'],protected=1,sequence=1,doc="Info showing status transitions and unit info"),
        'id': SimpleItem(defvalue=-1, protected=1, doc='ID of the Unit', typelist=["int"]),
    })

    _category = 'units'
    _name = 'IUnit'
    _exportmethods = []
    _hidden = 0

# Special methods:
    def __init__(self):
        super(IUnit, self).__init__()
        self.updateStatus("new")

    def _readonly(self):
        """A unit is read-only if the status is not new."""
        if self.status == "new":
            return 0
        return 1

    def validate(self):
        """Validate that this unit is OK and set it to active"""
        self.active = True
        return True

    def getID(self):
        """Get the ID of this unit within the transform"""

        # if the id isn't already set, use the index from the parent Task
        if self.id < 0:
           trf = self._getParent()
           if not trf:
              raise ApplicationConfigurationError(
                 None, "This unit has not been associated with a transform and so there is no ID available")
           self.id = trf.units.index(self)
           
        return self.id

    def updateStatus(self, status):
        """Update status hook"""
        addInfoString(self, "Status change from '%s' to '%s'" % (self.status, status))
        self.status = status

    def createNewJob(self):
        """Create any jobs required for this unit"""
        pass

    def checkCompleted(self, job):
        """Check if this unit is complete"""
        if job.status == "completed":
            return True
        else:
            return False

    def checkForSubmission(self):
        """Check if this unit should submit a job"""

        # check the delay
        if time.time() < self.start_time:
            return False

        # check if we already have a job
        if len(self.active_job_ids) != 0:
            return False

        # if we're using threads, check the max number
        if self._getParent().submit_with_threads and GPI.queues.totalNumUserThreads() > self._getParent().max_active_threads:
            return False

        return True

    def checkForResubmission(self):
        """check if this unit should be resubmitted"""

        # check if we already have a job
        if len(self.active_job_ids) == 0:
            return False
        else:
            job = GPI.jobs(self.active_job_ids[0])
            if job.status in ["failed", "killed"]:
                return True

            return False

    def checkParentUnitsAreComplete(self):
        """Check to see if the parent units are complete"""
        req_ok = True
        task = self._getParent()._getParent()
        for req in self.req_units:
            req_trf_id = int(req.split(":")[0])

            if req.find("ALL") == -1:
                req_unit_id = int(req.split(":")[1])
                if task.transforms[req_trf_id].units[req_unit_id].status != "completed":
                    req_ok = False

            else:
                # need all units from this trf
                for u in task.transforms[req_trf_id].units:
                    if u.status != "completed":
                        req_ok = False

        return req_ok

    def checkMajorResubmit(self, job):
        """check if this job needs to be fully rebrokered or not"""
        pass

    def majorResubmit(self, job):
        """perform a mjor resubmit/rebroker"""
        self.prev_job_ids.append(job.id)
        self.active_job_ids.remove(job.id)

    def minorResubmit(self, job):
        """perform just a minor resubmit"""
        try:
            trf = self._getParent()
        except Exception as err:
            logger.debug("GetParent exception!\n%s" % str(err))
            trf = None
        if trf is not None and trf.submit_with_threads:
            addInfoString( self, "Attempting job re-submission with queues..." )
            GPI.queues.add(job.resubmit)
        else:
            addInfoString( self, "Attempting job re-submission..." )
            job.resubmit()

    def update(self):
        """Update the unit and (re)submit jobs as required"""
        #logger.warning("Entered Unit %d update function..." % self.getID())

        # if we're complete, then just return
        if self.status in ["completed", "recreating"] or not self.active:
            return 0

        # check if submission is needed
        task = self._getParent()._getParent()
        trf = self._getParent()
        maxsub = task.n_tosub()

        # check parent unit(s)
        req_ok = self.checkParentUnitsAreComplete()

        # set the start time if not already set
        if len(self.req_units) > 0 and req_ok and self.start_time == 0:
            self.start_time = time.time() + trf.chain_delay * 60 - 1

        if req_ok and self.checkForSubmission() and maxsub > 0:

            # create job and submit
            addInfoString( self, "Creating Job..." )
            j = self.createNewJob()
            if j.name == '':
                j.name = "T%i:%i U%i" % (task.id, trf.getID(), self.getID())

            try:
                if trf.submit_with_threads:
                    addInfoString( self, "Attempting job submission with queues..." )
                    GPI.queues.add(j.submit)
                else:
                    addInfoString( self, "Attempting job submission..." )
                    j.submit()

            except Exception as err:
                logger.debug("update Err: %s" % str(err))
                addInfoString( self, "Failed Job Submission")
                addInfoString( self, "Reason: %s" % (formatTraceback()))
                logger.error("Couldn't submit the job. Deactivating unit.")
                self.prev_job_ids.append(j.id)
                self.active = False
                trf._setDirty()  # ensure everything's saved
                return 1

            self.active_job_ids.append(j.id)
            self.updateStatus("running")
            trf._setDirty()  # ensure everything's saved

            if trf.submit_with_threads:
                return 0

            return 1

        # update any active jobs
        for jid in self.active_job_ids:

            # we have an active job so see if this job is OK and resubmit if
            # not
            try:
                job = GPI.jobs(jid)
            except Exception as err:
                logger.debug("Update2 Err: %s" % str(err))
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            if job.status == "completed":

                # check if actually completed
                if not self.checkCompleted(job):
                    return 0

                # check for DS copy
                if trf.unit_copy_output:
                    if not self.copy_output:
                        trf.createUnitCopyOutputDS(self.getID())

                    if not self.copyOutput():
                        return 0

                # check for merger
                if trf.unit_merger:
                    if not self.merger:
                        self.merger = trf.createUnitMerger(self.getID())

                    if not self.merge():
                        return 0

                # all good so mark unit as completed
                self.updateStatus("completed")

            elif job.status == "failed" or job.status == "killed":

                # check for too many resubs
                if self.minor_resub_count + self.major_resub_count > trf.run_limit - 1:
                    logger.error("Too many resubmits (%i). Deactivating unit." % (
                        self.minor_resub_count + self.major_resub_count))
                    addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % ( self.minor_resub_count + self.major_resub_count))
                    self.active = False
                    return 0

                rebroker = False

                if self.minor_resub_count > trf.minor_run_limit - 1:
                    if self._getParent().rebroker_on_job_fail:
                        rebroker = True
                    else:
                        logger.error(
                            "Too many minor resubmits (%i). Deactivating unit." % self.minor_resub_count)
                        addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % (self.minor_resub_count + self.minor_resub_count))
                        self.active = False
                        return 0

                if self.major_resub_count > trf.major_run_limit - 1:
                    logger.error(
                        "Too many major resubmits (%i). Deactivating unit." % self.major_resub_count)
                    addInfoString( self, "Deactivating unit. Too many resubmits (%i)" % (self.minor_resub_count + self.major_resub_count))
                    self.active = False
                    return 0

                # check the type of resubmit
                if rebroker or self.checkMajorResubmit(job):

                    self.major_resub_count += 1
                    self.minor_resub_count = 0

                    try:
                        addInfoString( self, "Attempting major resubmit...")
                        self.majorResubmit(job)
                    except Exception as err:
                        logger.debug("Update Err3: %s" % str(err))
                        logger.error("Couldn't resubmit the job. Deactivating unit.")
                        addInfoString( self, "Failed Job resubmission")
                        addInfoString( self, "Reason: %s" % (formatTraceback()))
                        self.active = False

                    # break the loop now because we've probably changed the
                    # active jobs list
                    return 1
                else:
                    self.minor_resub_count += 1
                    try:
                        addInfoString( self, "Attempting minor resubmit...")
                        self.minorResubmit(job)
                    except Exception as err:
                        logger.debug("Update Err4: %s" % str(err))
                        logger.error("Couldn't resubmit the job. Deactivating unit.")
                        addInfoString( self, "Failed Job resubmission")
                        addInfoString( self, "Reason: %s" % (formatTraceback()))
                        self.active = False
                        return 1

    def reset(self):
        """Reset the unit completely"""
        addInfoString( self, "Reseting Unit...")
        self.minor_resub_count = 0
        self.major_resub_count = 0
        if len(self.active_job_ids) > 0:
            self.prev_job_ids += self.active_job_ids
        self.active_job_ids = []

        self.active = True

        # if has parents, set to recreate
        if len(self.req_units) > 0:
            self.updateStatus("recreating")
        else:
            self.updateStatus("running")

    # Info routines
    def n_active(self):

        if self.status == 'completed':
            return 0

        tot_active = 0
        active_states = ['submitted', 'running']

        for jid in self.active_job_ids:

            try:
                job = GPI.jobs(jid)
            except Exception as err:
                logger.debug("n_active Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) > 0:
                    for sj_stat in j.getNodeIndexCache()['subjobs:status']:
                        if sj_stat in active_states:
                            tot_active += 1
                else:
                    if j.getNodeIndexCache()['status'] in active_states:
                        tot_active += 1
            else:
                #logger.warning("WARNING: (active check) No index cache for job object %d" % jid)
                if j.status in active_states:
                    if j.subjobs:
                        for sj in j.subjobs:
                            if sj.status in active_states:
                                tot_active += 1
                    else:
                        tot_active += 1

        return tot_active

    def n_status(self, status):
        tot_active = 0
        for jid in self.active_job_ids:

            try:
                job = GPI.jobs(jid)
            except Exception as err:
                logger.debug("n_status Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) > 0:
                    for sj_stat in j.getNodeIndexCache()['subjobs:status']:
                        if sj_stat == status:
                            tot_active += 1
                else:
                    if j.getNodeIndexCache()['status'] == status:
                        tot_active += 1

            else:
                #logger.warning("WARNING: (status check) No index cache for job object %d" % jid)
                if j.subjobs:
                    for sj in j.subjobs:
                        if sj.status == status:
                            tot_active += 1
                else:
                    if j.status == status:
                        tot_active += 1

        return tot_active

    def n_all(self):
        total = 0
        for jid in self.active_job_ids:

            try:
                job = GPI.jobs(jid)
            except Exception as err:
                logger.debug("n_all Err: %s" % str(err))
                task = self._getParent()._getParent()
                trf = self._getParent()
                logger.warning("Cannot find job with id %d. Maybe reset this unit with: tasks(%d).transforms[%d].resetUnit(%d)" %
                               (jid, task.id, trf.getID(), self.getID()))
                continue

            j = stripProxy(job)

            # try to preserve lazy loading
            if hasattr(j, 'getNodeIndexCache') and j.getNodeIndexCache() and 'subjobs:status' in j.getNodeIndexCache():
                if len(j.getNodeIndexCache()['subjobs:status']) != 0:
                    total += len(j.getNodeIndexCache()['subjobs:status'])
                else:
                    total += 1
            else:
                #logger.warning("WARNING: (status check) No index cache for job object %d" % jid)
                if j.subjobs:
                    total = len(j.subjobs)
                else:
                    total = 1

        return total

    def overview(self):
        """Print an overview of this unit"""
        o = "    Unit %d: %s        " % (self.getID(), self.name)

        for s in ["submitted", "running", "completed", "failed", "unknown"]:
            o += markup("%i   " % self.n_status(s), overview_colours[s])

        print(o)

    def copyOutput(self):
        """Copy any output to the given dataset"""
        logger.error(
            "No default implementation for Copy Output - contact plugin developers")
        return False



import os
import base64
import subprocess
import threading
import pickle
import signal
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Credentials import getCredential
import Ganga.Utility.execute as gexecute


import time
import math
import copy

logger = getLogger()
proxy = getCredential('GridProxy', '')

# Cache
# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
DIRAC_ENV = {}
DIRAC_INCLUDE = ''
Dirac_Env_Lock = threading.Lock()
Dirac_Proxy_Lock = threading.Lock()

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def getDiracEnv(force=False):
    global DIRAC_ENV
    global Dirac_Env_Lock
    lock = Dirac_Env_Lock
    lock.acquire()
    if DIRAC_ENV == {} or force:
        if getConfig('DIRAC')['DiracEnvFile'] != "" and os.path.exists(getConfig('DIRAC')['DiracEnvFile']):
            with open(getConfig('DIRAC')['DiracEnvFile'], 'r') as env_file:
                DIRAC_ENV = dict((tuple(line.strip().split('=', 1)) for line in env_file.readlines(
                ) if len(line.strip().split('=', 1)) == 2))
                keys_to_remove = []
                for k, v in DIRAC_ENV.iteritems():
                    if str(v).startswith('() {'):
                        keys_to_remove.append(k)
                for key in keys_to_remove:
                    del DIRAC_ENV[key]

        else:
            logger.error(
                "'DiracEnvFile' config variable empty or file not present")
    lock.release()
    return DIRAC_ENV

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def getDiracCommandIncludes(force=False):
    global DIRAC_INCLUDE
    if DIRAC_INCLUDE == '' or force:
        for fname in getConfig('DIRAC')['DiracCommandFiles']:
            if not os.path.exists(fname):
                raise GangaException(
                    "Specified Dirac command file '%s' does not exist." % fname)
            with open(fname, 'r') as inc_file:
                DIRAC_INCLUDE += inc_file.read() + '\n'

    return DIRAC_INCLUDE

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


def getValidDiracFiles(job, names=None):
    from GangaDirac.Lib.Files.DiracFile import DiracFile
    from Ganga.GPIDev.Base.Proxy import isType
    if job.subjobs:
        for sj in job.subjobs:
            for df in (f for f in sj.outputfiles if isType(f, DiracFile)):
                if df.subfiles:
                    for valid_sf in (sf for sf in df.subfiles if sf.lfn != '' and (names is None or sf.namePattern in names)):
                        yield valid_sf
                else:
                    if df.lfn != '' and (names is None or df.namePattern in names):
                        yield df
    else:
        for df in (f for f in job.outputfiles if isType(f, DiracFile)):
            if df.subfiles:
                for valid_sf in (sf for sf in df.subfiles if sf.lfn != '' and (names is None or sf.namePattern in names)):
                    yield valid_sf
            else:
                if df.lfn != '' and (names is None or df.namePattern in names):
                    yield df

last_modified_time = None
last_modified_valid = False

# This will move/change when new credential system in place
############################


def _dirac_check_proxy( renew = True):
    global last_modified_valid
    global proxy
    _isValid = proxy.isValid()
    if not _isValid:
        if renew is True:
            proxy.renew()
            if not proxy.isValid():
                last_modified_valid = False
                raise GangaException('Can not execute DIRAC API code w/o a valid grid proxy.')
            else:
                last_modified_valid = True
        else:
            last_modified_valid = False
    else:
        last_modified_valid = True
############################

# /\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

def _proxyValid():
    _checkProxy( renew = False )
    return last_modified_valid

def _checkProxy( delay=60, renew = True ):
    ## Handling mutable globals in a multi-threaded system to remember to LOCK
    global Dirac_Proxy_Lock
    Dirac_Proxy_Lock.acquire()
    ## Function to check for a valid proxy once every 60( or n) seconds
    global last_modified_time
    if last_modified_time is None:
        # This will move/change when new credential system in place
        ############################
        _dirac_check_proxy( True )
        ############################
        last_modified_time = time.time()

    if abs(last_modified_time - time.time()) > int(delay):
        _dirac_check_proxy( renew )
        last_modified_time = time.time()

    Dirac_Proxy_Lock.release()

def execute(command,
            timeout=getConfig('DIRAC')['Timeout'],
            env=None,
            cwd=None,
            shell=False,
            python_setup='',
            eval_includes=None,
            update_env=False):
    """
    Execute a command on the local DIRAC server.

    This function blocks until the server returns.
    """

    if env is None:
        env = getDiracEnv()
    if python_setup == '':
        python_setup = getDiracCommandIncludes()

    _checkProxy()

    #logger.debug("Executing command:\n'%s'" % str(command))
    #logger.debug("python_setup:\n'%s'" % str(python_setup))
    #logger.debug("eval_includes:\n'%s'" % str(eval_includes))

    returnable = gexecute.execute(command,
                                  timeout=timeout,
                                  env=env,
                                  cwd=cwd,
                                  shell=shell,
                                  python_setup=python_setup,
                                  eval_includes=eval_includes,
                                  update_env=update_env)

    # rcurrie I've seen problems with just returning this raw object,
    # expanding it to be sure that an instance remains in memory
    myObject = {}
    if hasattr(returnable, 'keys'):
        # Expand object(s) in dictionaries
        myObject = _expand_object(returnable)
    elif type(returnable) in[list, tuple]:
        # Expand object(s) in lists
        myObject = _expand_list(returnable)
    else:
        # Copy object(s) so thet they definately are in memory
        myObject = copy.deepcopy(returnable)

    return myObject


def _expand_object(myobj):
    new_obj = {}
    if hasattr(myobj, 'keys'):
        for key in myobj.keys():
            value = myobj.get(key)
            if hasattr(value, 'keys'):
                new_obj[key] = _expand_object(value)
            else:
                new_obj[key] = copy.deepcopy(value)
    return new_obj


def _expand_list(mylist):
    new_list = []
    for element in mylist:
        new_list.append(copy.deepcopy(element))
    return new_list




# File: GangaAtlas/__init__.py

## Config options
from Ganga.Utility.Config import makeConfig, ConfigError, getConfig
import os

from Ganga.Utility.Config.Config import _after_bootstrap
from Ganga.Utility.logging import getLogger
logger = getLogger()

if not _after_bootstrap:

    # -------------------------------------------------
    # Athena Options
    config = makeConfig('Athena','Athena configuration parameters')
    config.addOption('LCGOutputLocation', 'srm://srm-atlas.cern.ch/castor/cern.ch/grid/atlas/scratch/%s/ganga' % os.environ['USER'], 'FIXME')
    config.addOption('LocalOutputLocation', '/castor/cern.ch/atlas/scratch/%s/ganga' % os.environ['USER'], 'FIXME')
    config.addOption('IndividualSubjobDirsForLocalOutput', False, 'When copying local output, should dir structure be jid.sid (False) or jid/sid (True)' )
    config.addOption('SingleDirForLocalOutput', False, 'When copying local output, only a single dirs used for output and output filenames are changed with jid.sid' )
    config.addOption('NoSubDirsAtAllForLocalOutput', False, 'When copying local output, all output is copied to the given output location with no subdirs created' )
    config.addOption('ATLAS_SOFTWARE', '/afs/cern.ch/project/gd/apps/atlas/slc3/software', 'FIXME')
    config.addOption('PRODUCTION_ARCHIVE_BASEURL', 'http://atlas-computing.web.cern.ch/atlas-computing/links/kitsDirectory/Production/kits/', 'FIXME')
    config.addOption('ExcludedSites', '' , 'FIXME')
    config.addOption('CMTHOME', os.path.join(os.environ['HOME'],'cmthome') , 'The path in which the cmtsetup magic function will look up the setup.sh for CMT environment setup')
    config.addOption('CMTCONFIG', 'i686-slc5-gcc43-opt', 'Default value to be used as CMTCONFIG environment setup value (LCG/Batch backend)')
    config.addOption('CMTCONFIG_LIST', [ 'i686-slc4-gcc34-opt', 'i686-slc5-gcc43-opt' ], 'Allowed values for CMTCONFIG environment setup value (LCG/Batch backend)')
    config.addOption('MaxJobsAthenaSplitterJobLCG', 1000 , 'Number of maximum jobs allowed for job splitting with the AthenaSplitterJob and the LCG backend')
    config.addOption('DCACHE_RA_BUFFER', 32768 , 'Size of the dCache read ahead buffer used for dcap input file reading')
    config.addOption('ENABLE_DQ2COPY', False , 'Enable DQ2_COPY input workflow on LCG backend')
    config.addOption('ENABLE_SGE_DQ2JOBSPLITTER', False , 'Enable DQ2JobSplitter for SGE backend')
    config.addOption('ENABLE_SGE_FILESTAGER', False , 'Enable FILE_STAGER input access mode for SGE backend')
    config.addOption('EXE_MAXFILESIZE', 1024*1024 , 'Athena.exetype=EXE jobs: Maximum size of files to be sent to WNs (default 1024*1024B)')
    config.addOption('dereferenceSymLinks', False , 'Set to True to dereference symlinks in the sources area. E.g. if src is a symlink, the target will be copied into the sources archive.')
    config.addOption('MaxJobsDQ2JobSplitter', 5000, 'Maximum number of allowed subjobs of DQ2JobSplitter')
    config.addOption('MaxFilesPandaDQ2JobSplitter', 5000, 'Maximum number of allowed subjobs of DQ2JobSplitter')
    config.addOption('MaxJobsDQ2JobSplitterLCGCompile', 500, 'Maximum number of allowed subjobs of DQ2JobSplitter on LCG/Cream with compile option switched on')
    config.addOption('MaxFileSizeNGDQ2JobSplitter', 14336, 'Maximum total sum of filesizes per subjob of DQ2JobSplitter at the NG backend (in MB)')
    config.addOption('MaxFileSizePandaDQ2JobSplitter', 13336, 'Maximum total sum of filesizes per subjob of DQ2JobSplitter at the Panda backend (in MB)')
    config.addOption('DefaultNumFilesPandaDirectDQ2JobSplitter', 50, 'Default number of input files per subjob used at a direct access site in Panda in DQ2JobSplitter')
    config.addOption('AllowedSitesNGDQ2JobSplitter', [ 'NDGF-T1_DATADISK', 'NDGF-T1_MCDISK', 'NDGF-T1_PRODDISK', 'NDGF-T1_SCRATCHDISK' ], 'Allowed space tokens/sites for DQ2JobSplitter on NG backend' )
    config.addOption('AnyCloudPreferenceList', [ ], 'List of clouds that should be preferentially submitted to when using the anyCloud option' )
    config.addOption('ATLASOutputDatasetLFC', 'prod-lfc-atlas-local.cern.ch', 'FIXME')
    config.addOption('PathToEOSBinary', '/afs/cern.ch/project/eos/installation/pro/bin/eos.select', 'Path to the EOS binary for output copying/checking')
    config.addOption('RemoveTempUserAreaAfterPrepare', True, 'If True, user areas created in /tmp are removed after prepare has been called (the version in the gangadir shared area will be used instead)')

    # -------------------------------------------------
    # AthenaMC Options
    config = makeConfig('AthenaMC', 'AthenaMC configuration options')

    # -------------------------------------------------
    # Athen RT Options
    mc = getConfig('MonitoringServices')
    mc.addOption('Athena/LCG', 'Ganga.Lib.MonitoringServices.ARDADashboard.LCG.ARDADashboardLCGAthena.ARDADashboardLCGAthena,Ganga.Lib.MonitoringServices.MSGMS.MSGMS,Ganga.Lib.MonitoringServices.Dashboard.LCGAthenaMS.LCGAthenaMS', 'sets job monitoring service for Athena/LCG jobs')
    mc.addOption('Athena/CREAM', 'Ganga.Lib.MonitoringServices.ARDADashboard.LCG.ARDADashboardLCGAthena.ARDADashboardLCGAthena,Ganga.Lib.MonitoringServices.MSGMS.MSGMS,Ganga.Lib.MonitoringServices.Dashboard.LCGAthenaMS.LCGAthenaMS', 'sets job monitoring service for Athena/CREAM jobs')
    mc.addOption('Athena', None, 'FIXME')

    # -------------------------------------------------
    # DQ2 Options
    config = makeConfig('DQ2', 'DQ2 configuration options')

    try:
        config.addOption('DQ2_URL_SERVER', os.environ['DQ2_URL_SERVER'], 'FIXME')
    except KeyError:
        config.addOption('DQ2_URL_SERVER', 'http://atlddmcat.cern.ch/dq2/', 'FIXME')
    try:
        config.addOption('DQ2_URL_SERVER_SSL', os.environ['DQ2_URL_SERVER_SSL'], 'FIXME')
    except KeyError:
        config.addOption('DQ2_URL_SERVER_SSL', 'https://atlddmcat.cern.ch:443/dq2/', 'FIXME')

    try:
        config.addOption('DQ2_LOCAL_SITE_ID', os.environ['DQ2_LOCAL_SITE_ID'], 'Sets the DQ2 local site id')
    except KeyError:
        config.addOption('DQ2_LOCAL_SITE_ID', 'CERN-PROD_DATADISK', 'Sets the DQ2 local site id')

    config.addOption('DQ2_OUTPUT_SPACE_TOKENS', [ 'ATLASSCRATCHDISK', 'ATLASLOCALGROUPDISK', 'T2ATLASSCRATCHDISK', 'T2ATLASLOCALGROUPDISK' ] , 'Allowed space tokens names of DQ2OutputDataset output' )

    config.addOption('DQ2_BACKUP_OUTPUT_LOCATIONS', [ 'CERN-PROD_SCRATCHDISK', 'CERN-PROD_USERTAPE', 'FZK-LCG2_SCRATCHDISK', 'IN2P3-CC_SCRATCHDISK', 'TRIUMF-LCG2_SCRATCHDISK', 'IFAE_SCRATCHDISK', 'NIKHEF-ELPROD_SCRATCHDISK' ], 'Default backup locations of DQ2OutputDataset output' )

    config.addOption('USE_STAGEOUT_SUBSCRIPTION', False, 'Allow DQ2 subscription to aggregate DQ2OutputDataset output on a storage element instead of using remote lcg-cr' )

    config.addOption('usertag','user','user tag for a given data taking period')

    config.addOption('USE_ACCESS_INFO', False, 'Use automatic best choice of input dataset access mode provided by AtlasLCGRequirements.')

    config.addOption('CHECK_OUTPUT_DUPLICATES', False, 'Check for duplicate files in DQ2OutputDataset in LCG backend - this could possibly happen by ShallowRetry of glite WMS. A duplicates dataset is created')
    config.addOption('DELETE_DUPLICATES_DATASET', False, 'If CHECK_OUTPUT_DUPLICATES=True is used, duplicates dataset can be automatically deleted by setting this flag to True.')

    config.addOption('USE_NICKNAME_DQ2OUTPUTDATASET', True, 'Use voms nicknames for DQ2OutputDataset.')
    config.addOption('ALLOW_MISSING_NICKNAME_DQ2OUTPUTDATASET', False, 'Allow that voms nickname is empty for DQ2OutputDataset name creating.')

    config.addOption('OUTPUTDATASET_LIFETIME', '', 'Maximum lifetime of a DQ2OutputDataset.')
    config.addOption('OUTPUTDATASET_NAMELENGTH', 131, 'Maximum characters of a DQ2OutputDataset.')
    config.addOption('OUTPUTFILE_NAMELENGTH', 150, 'Maximum characters of a filename in DQ2OutputDataset.')
    config.addOption('NumberOfDQ2DownloadThreads', 5, 'Number of simultaneous DQ2 downloads when calling "retrieve"')

    config.addOption('setupScript', '/cvmfs/atlas.cern.ch/repo/ATLASLocalRootBase/user/gangaDDMSetup.sh', 'Script to setup DQ2Clients software')

    # -------------------------------------------------
    # AMI Dataset Options
    config = makeConfig('AMIDataset','AMI dataset')
    config.addOption('MaxNumOfFiles', 3000, 'Maximum number of files in a given dataset patterns')
    config.addOption('MaxNumOfDatasets', 100, 'Maximum number of datasets in a given dataset patterns')

    # -------------------------------------------------
    # Tasks Options
    config = getConfig("Tasks")
    config.addOption('cloudPreference',[],'list of preferred clouds to choose for AnaTask analysis')
    config.addOption('backendPreference',["LCG","Panda","NG"],'order of preferred backends (LCG, Panda, NG) for AnaTask analysis')
    config.addOption('merged_files_per_job',1,'OBSOLETE', type=int)
    config.addOption('recon_files_per_job',10,'OBSOLETE', type=int)



def loadPlugins( config = {} ):

   import warnings
   warnings.filterwarnings('ignore','Python C API version mismatch for module pycurl')
   warnings.filterwarnings('ignore','Python C API version mismatch for module _lfc')

   import Lib.Athena
   import Lib.ATLASDataset
   import Lib.AthenaMC
   import Lib.Tnt
   import Lib.AtlasLCGRequirements
   import Lib.Tasks
   import Lib.TagPrepare
   import Lib.AMIGA
   
   return None

def getEnvironment(c):
    import PACKAGE
    PACKAGE.standardSetup()
    
    #   set up X509_CERT_DIR for DQ2
    from Ganga.Utility.GridShell import getShell
    gshell = getShell()
    if gshell:
       try:
          return { 'X509_CERT_DIR' : gshell.env['X509_CERT_DIR'], 'X509_USER_PROXY' : gshell.env['X509_USER_PROXY']  }
       except KeyError:
          return { 'X509_CERT_DIR' : '/etc/grid-security/certificates' }


# some checks to make sure that new Dashboard MSG service is enabled in the user's configuration

def postBootstrapHook():
    from Ganga.Utility.logging import getLogger

    logger = getLogger()

    from Ganga.Utility.Config import getConfig
    cfg = getConfig('MonitoringServices')

    MONITORING_DEFAULT = "Ganga.Lib.MonitoringServices.Dashboard.LCGAthenaMS.LCGAthenaMS"

    for name in cfg.options:
        value = cfg[name]
        if 'Athena' in name.split('/') and ('LCG' in name.split('/') or 'CREAM' in name.split('/')):
            if not MONITORING_DEFAULT in value.split(','):
             logger.error('''*** Outdated monitoring configuration - check your configuration files ***

*** Outdated monitoring configuration - check your configuration files ***

Maybe your ~/.gangarc contains old entries which override new defaults?
You may also check the configuration files defined by $GANGA_CONFIG_PATH or $GANGA_CONFIG environment variables.

To fix this problem simply remove (or comment out) the following lines in [MonitoringServices] section:
Athena/LCG=...
Athena/CREAM=...

For now I will add the correct default settings (%s) to the configuration of this Ganga session.
Note that in the future you won't be able to start Ganga until these issues are corrected manually.
'''%MONITORING_DEFAULT)

             cfg.setUserValue(name,value+','+MONITORING_DEFAULT)





#from TaskApplication import *
from LHCbAnalysisTransform import LHCbAnalysisTransform
from LHCbAnalysisTask import LHCbAnalysisTask
#from LHCbTaskDummySplitter import *
from BKTestQuery import BKTestQuery

from LHCbTask import LHCbTask
from LHCbTransform import LHCbTransform
from LHCbUnit import LHCbUnit



#
# Splitter for TNT plugin
#

import sys, os, re
import popen2
import signal
import getopt, string, time 
import shutil

from Ganga.GPIDev.Base import *
from Ganga.GPIDev.Schema import *
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.Lib.Executable import Executable
from Ganga.GPIDev.Lib.Job import *
from Ganga.GPIDev.Lib.File import *
from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import *
from Ganga.Core.exceptions import ApplicationConfigurationError

_subjobs = []
_subCollections = {} 
_guidLfnMap = {}
_guidDatasetMap = {}
_strippedCollection = ''
_tagfileCollection = ''

class TNTJobSplitter(ISplitter):
   
   _name = "TNTJobSplitter"
   _exportmethods = ['tagNavigatorTool','runQuery']
   _schema = Schema(Version(1,0), { 
      'src_collection_name'   : SimpleItem(defvalue='', doc='Source collection name'),
      'src_collection_type'   : SimpleItem(defvalue='RootCollection', doc='Source collection type'),
      'src_connection_string' : SimpleItem(defvalue='', doc='Source connection string'),
      'query'                 : SimpleItem(defvalue='', doc='Database query to execute'),
      'logfile'               : SimpleItem(defvalue='TNT.log-'+ str(os.getpid()),doc='Logfile name'),
      'auth_name'             : SimpleItem(defvalue='', doc='Username from authentication.xml'),
      'auth_pass'             : SimpleItem(defvalue='', doc='Password for username from authentication.xml'),
      #'output_collection'     : SimpleItem(defvalue='myEvents.root', doc='Name of query output collection'),
      #'collection': SimpleItem(defvalue='myEvents.root', doc="Collection you want to process"),
      'minevents' : SimpleItem(defvalue='0', doc='Minimum number of events per sub collection'),
      'match_ce'  : SimpleItem(defvalue=False, doc='Run jobs on CEs with local dataset')
      } )

   # method to add date and time to every print statement
   def tntPrint(self, textForPrinting):
         #open the logfile
         writefile = open(self.inDir + "/" + self.logfile,"a")
         writefile.write("[" + time.strftime('%a %b %d %H:%M:%S') + "]  " + textForPrinting +"\n")
         logger.info(textForPrinting)
         writefile.close
  
   def writeAuth(self):
      authfile = open("authentication.xml","w")
      authfile.write("<connectionlist>\n")
      authfile.write("<connection name=\"" + self.src_connection_string + "\">\n")
      authfile.write("<parameter name=\"user\" value=\"" + self.auth_name + "\"/>\n")
      authfile.write("<parameter name=\"password\" value=\"" + self.auth_pass + "\"/>\n")
      authfile.write("</connection>\n")
      authfile.write("</connectionlist>\n")

   def writeCollAppendXML(self):
      collAppendfile = open("CollAppend.exe.xml","w")
      collAppendfile.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n")
      collAppendfile.write("<!DOCTYPE ArgList>\n")
      collAppendfile.write("<ArgList>\n")
      collAppendfile.write("<ToolInfo date=\"Thu Nov 20 18:29:03 2008&#xA;\" toolID=\"CollAppend.exe\">\n")
      collAppendfile.write("<CliArg name=\"General\" option=\"-ignoreschemadiff\"/>\n")
      collAppendfile.write("<CliArg name=\"General\" option=\"-nevents\">1</CliArg>\n")
      collAppendfile.write("<CliArg name=\"General\" option=\"-nevtcached\">1000000</CliArg>\n")
      collAppendfile.write("<CliArg name=\"General\" option=\"-nevtperprint\">100</CliArg>\n")
      collAppendfile.write("<CliArg name=\"General\" option=\"-noattrib\"/>\n")
      #collAppendfile.write("<CliArg name=\"QueryInfo\" option=\"-queryopt\">RunNumber, EventNumber, StreamAOD_ref</CliArg>\n")
      collAppendfile.write("<CliArg name=\"DstInfo\" option=\"-dst\">"+self.inDir + self.stripped_coll_name + " RootCollection"+"</CliArg>\n")
      collAppendfile.write("<CliArg name=\"DstInfo\" option=\"-dstconnect\">none</CliArg>\n")
      collAppendfile.write("<CliArg name=\"SrcInfo\" option=\"-src\">"+self.src_collection_name+" RootCollection"+"</CliArg>\n")
      collAppendfile.write("</ToolInfo>\n")
      collAppendfile.write("</ArgList>\n")


   def writeCollSplitByGUIDXML(self):
      collSplitfile = open("CollSplitByGUID.exe.xml","w")
      collSplitfile.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n")
      collSplitfile.write("<!DOCTYPE ArgList>\n")
      collSplitfile.write("<ArgList>\n")
      collSplitfile.write("<ToolInfo date=\"Thu Nov 20 18:29:03 2008&#xA;\" toolID=\"CollSplitByGUID.exe\">\n")
      collSplitfile.write("<CliArg name=\"General\" option=\"-minevents\">0</CliArg>\n")
      #collSplitfile.write("<CliArg name=\"General\" option=\"-splitref\">StreamESD</CliArg>\n")        
      collSplitfile.write("<CliArg name=\"SrcInfo\" option=\"-src\">"+self.src_collection_name+" RootCollection"+"</CliArg>\n")
      collSplitfile.write("</ToolInfo>\n")
      collSplitfile.write("</ArgList>\n") 
       
   def runQuery(self):
      ### ---------------------  initialisation -------------------------------------------
      jobObj = self.getJobObject()
      self.inDir = jobObj.getInputWorkspace().getPath()
      self.output_collection = 'myEvents.root'   # fix these instead of having them variable - too confusing
      self.stripped_coll_name = re.split('\.root', self.output_collection)[0]
      #print self.stripped_coll_name
      self.collection = 'myEvents.root'
      self.tntPrint("**** Welcome to TNT! ****")
      self.tntPrint("Using " + str(self.inDir) + " to store intermediary files")
      self.tntPrint(self.logfile)
      #does authentication.xml exist in the current directory?
      cwd = os.getcwd()
      if not os.path.exists("authentication.xml"):
         self.tntPrint("authentication.xml not found in current directory ("+cwd+")...creating file.")
         self.writeAuth()
      else:
         self.tntPrint("authentication.xml found in " + cwd)
   
      #does CollAppend.exe..xml exist in the current directory?
      cwd = os.getcwd()
      if not os.path.exists("CollAppend.exe..xml"):
         self.tntPrint("CollAppend.exe.xml not found in current directory ("+cwd+")...creating new file.")
         self.writeCollAppendXML()
      else:
         self.tntPrint("CollAppend.exe.xml found -  overwriting old version " + cwd)      
         self.writeCollAppendXML()

      #does CollAppend.exe..xml exist in the current directory?
      cwd = os.getcwd()
      if not os.path.exists("CollSplitByGUID.exe..xml"):
         self.tntPrint("CollSplitByGUID.exe.xml not found in current directory ("+cwd+")...creating new file.")
         self.writeCollSplitByGUIDXML()
      else:
         self.tntPrint("CollSplitByGUID.exe.xml found -  overwriting old version " + cwd)
         self.writeCollSplitByGUIDXML()      

      #remove existing files 
      #mjk not necessary now because all done in workspace?
      if os.path.exists(self.inDir + "/" + self.output_collection):
         os.remove(self.inDir + "/" + self.output_collection)
   
   
      ### ----------------------------- querying and extracting of GUIDs ----------------------------
      
      #stripped_coll_name = re.split('\.root', self.output_collection)[0]
      
      ### execute query and write events to myEvents.root
      self.tntPrint("Executing query '" + self.query + "' on tag database...")

      #prepare the CollAppend command for passing to the system.
      #collcommand="which CollAppend && CollAppend -src " + self.src_collection_name + " " + \
      #self.src_collection_type + " -srcconnect " + self.src_connection_string + \
      #" -dst " + self.inDir + stripped_coll_name + " RootCollection -queryopt \'RunNumber, EventNumber, StreamAOD_ref\' -query \"" + \
      #self.query + "\""
      collcommand="CollAppend.exe -xmlInput CollAppend.exe.xml"
      #print "########################"
      #print "CALLING CollAppend - but not really needed as not executing query and therefore not making collection - MyEvent"
      #print "########################"

      #pass CollAppend command to the system
      outputhandle, inputhandle = popen2.popen4(collcommand)
      inputhandle.close()
      output = outputhandle.readlines()
      #print output
      outputhandle.close()
      for line in output:
         self.tntPrint(line)
         
   def collSplit(self):
      """ splits an event collection into a number of sub-collections, 
          based on GUID boundaries but with a specified number of events per 
          sub-collection """
 
      self.tntPrint("TNTJobSplitter being called now")

      # split the collection into sub-collections
      masterCollection = re.split('\.root', self.collection)[0]
      #print "########################"
      #print "Master collection is"
      #print masterCollection
      #print "########################"
      global _strippedCollection
      _strippedCollection = masterCollection
      global _tagfileCollection
      _tagfileCollection = self.src_collection_name

      # make sure there is nothing already stored in collection list
      global _subCollections
      _subCollections = {}
  
      # TODO: when CollSplitByGUID in POOL release, use that instead of having local executable
      #exe = os.path.join(os.path.dirname(__file__),'CollSplitByGUID')
      #splitCommand=exe + " -src " + self.inDir + masterCollection + " RootCollection -minevents " + self.minevents 
      #print "########################"
      #print "CALLING CollSplitByGUID - Note using input tag file and not myEvent.root created from CollAppend"
      #print "########################" 
      splitCommand="CollSplitByGUID.exe -xmlInput CollSplitByGUID.exe.xml" 
      self.tntPrint(splitCommand)
      outputhandle, inputhandle = popen2.popen4(splitCommand)
      inputhandle.close()
      output = outputhandle.readlines()
      #print output
      outputhandle.close()
      for line in output:   
         collName = ""
         guidList = []
         if line.startswith("Created"):
            words = line.split()
            collName = words[3]
            self.tntPrint("Moving " + collName + ".root to " + self.inDir)
            #move sub-collections to Ganga workspace
            shutil.move(collName + ".root",self.inDir)
            for word in words[6:]:
               guidList.append(word)
            # collect mapping of sub-collections to the guids they contain
            _subCollections[self.inDir + collName] = guidList   


   def collInfo(self, collection):
      """ gets GUID, LFN and dataset info about given collection, using GuidExtractor """
      from GangaAtlas.Lib.Tnt import GuidExtractor
      self.tntPrint("About to call GuidExtractor")
      #GuidExtractor.pythonCalled(['p', collection])
      # NOTE USED TO PASS IN myEVENT above but no CollAppend used to put tag file straight in
      GuidExtractor.pythonCalled(['p', collection])
      global _guidLfnMap 
      _guidLfnMap = GuidExtractor.getGuidLfnMap()
      #print "guidLfnMap"
      #print _guidLfnMap
      #print "guidDatasetMap"
      global _guidDatasetMap 
      _guidDatasetMap = GuidExtractor.getGuidDatasetMap()
      #print _guidDatasetMap
      #move PoolFileCatalogues to Ganga workspace
      #if os.path.exists("PoolFileCatalog.xml"):
      #   shutil.move("PoolFileCatalog.xml",self.inDir)
   
   def _initSubJob(self, masterjob, dataset, lfnList, guidList, subCollection): 
       from Ganga.GPIDev.Lib.Job import Job
       from GangaAtlas.Lib.ATLASDataset import DQ2Dataset

       #print "########################"
       #print "SUBJOB INITIALISED"
       #print "########################"
       
       subjob = Job()
       subjob.inputsandbox = masterjob.inputsandbox
       subjob.application = masterjob.application
       subjob.outputdata = masterjob.outputdata
       subjob.outputsandbox = masterjob.outputsandbox
       subjob.backend = masterjob.backend

       # attributes which are different for each sub-job
       subjob.inputdata = DQ2Dataset()   
       #subjob.inputdata.datatype = 'DATA'
       subjob.inputdata.dataset = dataset
       #print "########################"
       #print "DATASET USED"
       #print dataset
       #print "########################"
       subjob.inputdata.names = lfnList
       #print "########################"
       #print "LFNLIST USED"
       #print lfnList
       #print "########################" 
       subjob.inputdata.guids = guidList
       #print "########################"
       #print "GUIDLIST USED"
       #print guidList
       #print "########################"
      
         
       if self.match_ce == True:
            #subjob.inputdata.type = 'TNT_LOCAL'

            # Sort out the possible sites taking into account requirements
            allowed_sites = []
            if subjob.backend.requirements.sites:
               allowed_sites = subjob.backend.requirements.sites
            elif subjob.backend.requirements.cloud:
               allowed_sites = subjob.backend.requirements.list_sites_cloud()
            else:
               raise ApplicationConfigurationError(None,'TntJobSplitter requires a cloud or a site to be set - please use the --cloud option, j.backend.requirements.cloud=CLOUDNAME (T0, IT, ES, FR, UK, DE, NL, TW, CA, US, NG) or j.backend.requirements.sites=SITENAME')

            if subjob.backend.requirements.sites:
               allowed_sites = subjob.backend.requirements.sites

            allowed_sites_all = subjob.backend.requirements.list_sites(True,True)
            # Apply GangaRobot blacklist
            newsites = []
            for site in allowed_sites:
               if site in allowed_sites_all:
                  newsites.append(site)
                  
            allowed_sites = newsites
                    
            # go through and check which sites with the dataset is given by the requirements
            sub_sites = []
            for site in subjob.inputdata.get_locations():
               if site in allowed_sites:
                  sub_sites.append(site)
                  
            if len(sub_sites) == 0:
               raise ApplicationConfigurationError(None,'TntJobSplitter could not find a location for dataset %s in cloud %s. Try another cloud!' % (subjob.inputdata.dataset, subjob.backend.requirements.cloud))
            else:
               subjob.backend.requirements.sites = sub_sites

            #print sub_sites
                        
       else:
            subjob.inputdata.type = 'TNT_DOWNLOAD'   #requires PFNs in sfn:// format

       subjob.inputsandbox += [ File(os.path.join(subCollection+".root")) ]
       
       return subjob

   def jobSplit(self, job):
      """ generates 1 sub-job per sub-collection """
      from Ganga.GPIDev.Lib.Job import Job
      from GangaAtlas.Lib.ATLASDataset import DQ2Dataset

      # make sure correct collection info is present
      self.tntPrint("Getting collection info...")
      self.tntPrint("Called guid extractor with " + str(self.inDir + _strippedCollection))
      #print "Called guid extractor with " 
      #print self.inDir + _strippedCollection
      #print _tagfileCollection
      # Want tag file passed in not stripped collection
      #self.collInfo(self.inDir + _strippedCollection)
      self.collInfo(_tagfileCollection)

      self.tntPrint("About to split job")
      #print "########################"
      #print "About to split job"
      #print "########################"
      
      subjobs = []
      for collection in _subCollections:
         guidList = _subCollections[collection]
         #print "########################"
         #print "This is the guidList "
         #print "########################"
         #print guidList 
         finalGuidList = []
         datasetList = []
         lfnList = []
         i = 0
         for guid in guidList:
            #print "########################"
            #print "This is the current guid in the list "
            #print guid
            #print "########################"  
            #print "########################"
            #print "This is the guid dataset map "
            #print _guidDatasetMap
            #print "########################"
            dataset = _guidDatasetMap[guid]
            if dataset != 'NO.DATASET':
               if datasetList.count(dataset) == 0:
                  datasetList.append(dataset)
                  i = i + 1
               finalGuidList.append(guid)
               lfn = _guidLfnMap[guid]
               lfnList.append(lfn)
         if i > 1:
            self.tntPrint("More than one dataset has been identified for a sub-job,")
            self.tntPrint("so sub-job has been copied into " + str(i) + " sub-jobs.")
            self.tntPrint("Correct behaviour of your analysis is not guaranteed in this case.")
            for dataset in datasetList:
                sj = self._initSubJob(job, dataset, lfnList, finalGuidList, collection)                        
                subjobs.append(sj)
         else:
             if dataset != 'NO.DATASET':
                sj = self._initSubJob(job, dataset, lfnList, finalGuidList, collection)           
                subjobs.append(sj)
      
      global _subjobs
      _subjobs = subjobs


   def split(self,job):
      #print "########################"
      #print "CALLING self.runQuery()"
      #print "########################"
      self.runQuery()
      #print "########################"
      #print "CALLING self.collsplit()"
      #print "########################"
      self.collSplit()
      #print "########################"
      #print "CALLING self.jobSplit()"
      #print "########################"
      self.jobSplit(job)
      return _subjobs


